"""Different utility functions for internal usage."""
import scipy.linalg.lapack as lapack
import pandas as pd
import numpy as np


def duplication_matrix(n: int):
    """
    Create duplication matrix D such that D vech(X) = vec(X).

    Parameters
    ----------
    n : int
        Size n of n x n matrix X.

    Returns
    -------
    Duplication matrix D.

    """
    m = n * n
    k = (n + 1) * n // 2
    d = np.zeros((m, k))
    for i in range(n):
        for j in range(i + 1):
            a1 =  i * n + j
            a2 =  j * n + i
            b = j * n + i - j * (j + 1) // 2
            d[a1, b] = 1
            d[a2, b] = 1
    return d

def kron_identity(mx: np.ndarray, sz: int, back=False):
    """
    Calculate Kronecker product with identity matrix.

    Simulates np.kron(mx, np.identity(sz)).
    Parameters
    ----------
    mx : np.ndarray
        Matrix.
    sz : int
        Size of identity matrix.
    back : bool, optional
        If True, np.kron(np.identity(sz), mx) will be calculated instead. The
        default is False.

    Returns
    -------
    np.ndarray
        Kronecker product of mx and an indeity matrix.

    """
    m, n = mx.shape
    r = np.arange(sz)
    if back:
        out = np.zeros((sz, m, sz, n), dtype=mx.dtype)
        out[r,:,r,:] = mx
    else:
        out = np.zeros((m, sz, n, sz), dtype=mx.dtype)
        out[:,r,:,r] = mx
    out.shape = (m * sz,n * sz)
    return out

def delete_mx(mx: np.ndarray, exclude: np.ndarray):
    """
    Remove column and rows from square matrix.

    Parameters
    ----------
    mx : np.ndarray
        Square matrix.
    exclude : np.ndarray
        List of indices corresponding to rows/cols.

    Returns
    -------
    np.ndarray
        Square matrix without certain rows and columns.

    """
    return np.delete(np.delete(mx, exclude, axis=0), exclude, axis=1)


def cov(x: np.ndarray):
    """
    Compute covariance matrix takin in account missing values.

    Parameters
    ----------
    x : np.ndarray
        Data.

    Returns
    -------
    np.ndarray
        Covariance matrix.

    """
    masked_x = np.ma.array(x, mask=np.isnan(x))
    cov = np.ma.cov(masked_x, bias=True, rowvar=False).data
    if cov.size == 1:
        cov.resize((1,1))
    return cov


def cor(x: np.ndarray):
    """
    Compute correlation matrix takin in account missing values.

    Parameters
    ----------
    x : np.ndarray
        Data.

    Returns
    -------
    np.ndarray
        Correlation matrix.

    """
    masked_x = np.ma.array(x, mask=np.isnan(x))
    cor = np.ma.corrcoef(masked_x, bias=True, rowvar=False).data
    if cor.size == 1:
        cor.resize((1,1))
    return cor

def chol(x: np.array, inv=True):
    """
    Calculate cholesky decomposition of matrix.

    Parameters
    ----------
    x : np.ndarray
        Matrix.
    inv : bool, optional
        If True, returns L^{-1} instead. The default i True.

    Returns
    -------
    Cholesky decomposition matrix L.

    """
    c, info = lapack.dpotrf(x)
    if inv:
        lapack.dtrtri(c, overwrite_c=1)
    return c
    


def chol_inv(x: np.array):
    """
    Calculate invserse of matrix using Cholesky decomposition.

    Parameters
    ----------
    x : np.array
        Data with columns as variables and rows as observations.

    Raises
    ------
    np.linalg.LinAlgError
        Rises when matrix is either ill-posed or not PD.

    Returns
    -------
    c : np.ndarray
        x^(-1).

    """
    c, info = lapack.dpotrf(x)
    if info:
        raise np.linalg.LinAlgError
    lapack.dpotri(c, overwrite_c=1)
    return c + c.T - np.diag(c.diagonal())


def chol_inv2(x: np.ndarray):
    """
    Calculate invserse and logdet of matrix using Cholesky decomposition.

    Parameters
    ----------
    x : np.ndarray
        Data with columns as variables and rows as observations.

    Raises
    ------
    np.linalg.LinAlgError
        Rises when matrix is either ill-posed or not PD.

    Returns
    -------
    c : np.ndarray
        x^(-1).
    logdet : float
        ln|x|

    """
    c, info = lapack.dpotrf(x)
    if info:
        raise np.linalg.LinAlgError
    d = c.diagonal()
    logdet = 2 * np.sum(np.log(d))
    lapack.dpotri(c, overwrite_c=1)
    return c + c.T - np.diag(d), logdet


def compare_results(model, true: pd.DataFrame, error='relative',
                    ignore_cov=True, drop_equal=True, mandatory=False,
                    return_table=False):
    """
    Compare parameter estimates in model to parameter values in a DataFrame.

    Parameters
    ----------
    model : Model
        Model instance.
    true : pd.DataFrame
        DataFrame with operations and expected estimates. Should have "lval",
        "op", "rval", "Value" columns in this particular order.
    error : str, optional
        If 'relative', relative errors are calculated. Absolute errors are
        calculated otherwise. The default is 'relative'.
    ignore_cov : bool, optional
        If True, then covariances (~~) are ignored. The default is True.
    drop_equal : bool, optional
        If True, then parameters that are exactly equal in model are dropped.
        Effectively, it clears result from fixed loadings in Lambda. The
        default is True.
    mandatory : bool, optional
        If True, then then all of entries in true DataFrame must have their
        counterparts in model. It might be a problem if there is a known
        variance estimate for an exogenous variable and you are testing
        ModelMeans/ModelEffects that lack variance parameter. The default is
        False.
    return_table : bool, optional
        If True, then pd.DataFrame table is returned instead of nplist.

    Raises
    ------
    Exception
        Rise when operation present in true is not present in the model.

    Returns
    -------
    errs : list
        List of errors.

    """
    if type(model) is not pd.DataFrame:
        ins = model.inspect(information=None)
    else:
        ins = model
    errs = list()
    if return_table:
        ops = list()
        lvals = list()
        rvals = list()
        ests = list()
        trues = list()
    for row in true.iterrows():
        lval, op, rval, value = row[1].values[:4]
        if op == '~~' and ignore_cov:
            continue
        if op == '=~':
            op = '~'
            lval, rval = rval, lval
        est = ins[(ins.lval == lval) & (ins.op == op) & (ins.rval == rval)]
        if len(est) == 0:
            if mandatory:
                raise Exception(f'Unknown estimate: {row}.')
            continue
        est = est.Estimate.values[0]
        if drop_equal and est == value:
            continue
        if error == 'relative':
            errs.append(abs((value - est) / est))
        else:
            errs.append(abs(value - est))
        if return_table:
            ops.append(op)
            lvals.append(lval)
            rvals.append(rval)
            ests.append(est)
            trues.append(value)
    if return_table:
        d = {'lval': lvals, 'op': ops, 'rval': rvals, 'Error': errs,
             'Estimate': ests, 'True': trues}
        df = pd.DataFrame.from_dict(d).sort_values('Error', axis=0, 
                                                   ascending=False)
        return df
    return errs


def calc_zkz(groups: pd.Series, k: pd.DataFrame, p_names=None, 
             return_z=False):
    """
    Calculate ZKZ^T relationship matrix from covariance matrix K.

    Parameters
    ----------
    groups : pd.Series
        Series of group names for individuals.
    k : pd.DataFrame
        Covariance-across-groups matrix. If None, then its calculate as an
        identity matrix.
    p_names : List[str], optional
        List of unique group names. If False, then it is determined from the
        groups variable. The default is None.
    return_z : bool, optional
        If True, then only Z is returned. The default is False.

    Raises
    ------
    Exception
        Incorrect number of groups: mismatch between dimensions of K and
        groups series.
    KeyError
        Incorrect group naming.

    Returns
    -------
    np.ndarray
        ZKZ^T matrix.

    """
    if p_names is None:
        lt = list()
        p_names = groups.unique()
        for gt in groups.unique():
            if type(gt) is str:
                lt.extend(gt.split(';'))
            else:
                if not np.isfinite(gt):
                    continue
                lt.append(gt)
        p_names = sorted(set(lt))
    p, n = len(p_names), len(groups)
    z = np.zeros((n, p))
    for i, germ in enumerate(groups):
        if type(germ) is str:
            for g in germ.split(';'):
                try:
                    j = p_names.index(g)
                    z[i, j] = 1.0
                except ValueError:
                    continue
        else:
            try:
                j = p_names.index(germ)
                z[i, j] = 1.0
            except ValueError:
                continue
    if return_z:
        return z
    if k is None:
        k = np.identity(p)
    elif k.shape[0] != p:
        raise Exception("Dimensions of K don't match number of groups.")
    if type(k) is pd.DataFrame:
        try:
            k = k.loc[p_names, p_names].values
        except KeyError:
            raise KeyError("Certain groups in K differ from those "
                           "provided in a dataset.")
    zkz = z @ k @ z.T
    return zkz


def calc_reduced_ml(model, variables: set, x=None,
                       exclude=False):
    """
    Calculate ML with restricted/reduced sigma.

    Parameters
    ----------
    model : Model
        DESCRIPTION.
    variables : set
        List/set of variable to be included in Sigma.
    x : np.ndarray, optional
        Paremters vector. The default is None.
    exclude : bool, optional
        If True, then variables are a list to be excluded, not included. The
        default is False.

    Returns
    -------
    float
        Degenerate ML.

    """
    
    from .model import Model
    if type(model) is not Model:
        raise Exception('ModelMeans or ModelEffects degenerate ML is not '
                        'supported.')
    obs = model.vars['observed']
    if exclude:
        inds = [obs.index(v) for v in variables]
    else:
        inds = [i for i, v in enumerate(obs) if v not in variables]
    def deg_sigma():
        sigma, (m, c) = true_sigma()
        sigma = delete_mx(sigma, inds)
        m = np.delete(m, inds, axis=0)
        return sigma, (m, c)
    true_sigma = model.calc_sigma
    true_cov = model.mx_cov
    true_covlogdet = model.cov_logdet
    model.mx_cov = delete_mx(true_cov, inds)
    model.cov_logdet = np.linalg.slogdet(model.mx_cov)[1]
    model.calc_sigma = deg_sigma
    if x is None:
        x = model.param_vals
    if type(model) is Model:
        ret = model.obj_mlw(x)
    else:
        ret = model.obj_fiml(x)
    model.calc_sigma = true_sigma
    model.mx_cov = true_cov
    model.cov_logdet = true_covlogdet
    return ret



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Inspector module helps researcher fetch information on estimates."""
from .optimizer import Optimizer
from .model import Model
import pandas as pd
import numpy as np
from . import stats


def inspect(model, mode='list', what='est', information='expected',
            std_est=False, se_robust=False):
    """
    Get fancy view of model parameters estimates.

    Parameters
    ----------
    model : str
        Model.
    mode : str, optional
        If 'list', pd.DataFrame with estimates and p-values is returned.
        If 'mx', a dictionary of matrices is returned. The default is 'list'.
    what : TYPE, optional
        Used only if mode == 'mx'. If 'est', matrices have estimated
        values. If 'start', matrices have starting values. If 'name', matrices
        have names inplace of their parameters. The default is 'est'.
    information : str
        If 'expected', expected Fisher information is used. Otherwise,
        observed information is employed. If None, the no p-values are
        calculated. No effect is what is 'mx'. The default is 'expected'.
    std_est : bool or str
        If True, standardized coefficients are also returned as Std. Ests col.
        If it is 'lv', then output variables are not standardized. The default
        is False.
    se_robust : bool, optional
        If True, then robust SE are computed instead. Robustness here means
        that MLR-esque sandwich correction is applied. The default is False.

    Returns
    -------
    pd.DataFrame | dict
        Dataframe or mapping matrix_name->matrix.

    """
    if isinstance(model, Optimizer):
        model = model.model
    if mode == 'list':
        return inspect_list(model, information=information, std_est=std_est,
                            se_robust=se_robust)
    elif mode == 'mx':
        return inspect_matrices(model, what=what)


def inspect_matrices(model: Model, what='est'):
    """
    Get model matrices.

    Parameters
    ----------
    model : Model
        Model.
    what : str, optional
        If 'est', matrices have estimated values. If 'start', matrices have
        starting values. If 'name', matrices have names inplace of their
        parameters. The default is 'est'.

    Raises
    ------
    Exception
        DESCRIPTION.

    Returns
    -------
    ret : TYPE
        DESCRIPTION.

    """
    ret = dict()
    if hasattr(model, 'mx_beta'):
        matrix = model.mx_beta
        name = 'Beta'
        names = model.names_beta
        if what != 'est':
            matrix = _set_values(model.parameters, matrix, what == 'start')
        ret[name] = pd.DataFrame(matrix, columns=names[1],
                                 index=names[0])

    if hasattr(model, 'mx_gamma1'):
        matrix = model.mx_gamma1
        name = 'Gamma1'
        names = model.names_gamma1
        if what != 'est':
            matrix = _set_values(model.parameters, matrix, what == 'start')
        ret[name] = pd.DataFrame(matrix, columns=names[1],
                                 index=names[0])

    if hasattr(model, 'mx_gamma2'):
        matrix = model.mx_gamma2
        name = 'Gamma2'
        names = model.names_gamma2
        if what != 'est':
            matrix = _set_values(model.parameters, matrix, what == 'start')
        ret[name] = pd.DataFrame(matrix, columns=names[1],
                                 index=names[0])

    if hasattr(model, 'mx_lambda'):
        matrix = model.mx_lambda
        name = 'Lambda'
        names = model.names_lambda
        if what != 'est':
            matrix = _set_values(model.parameters, matrix, what == 'start')
        ret[name] = pd.DataFrame(matrix, columns=names[1],
                                 index=names[0])
    if hasattr(model, 'mx_psi'):
        matrix = model.mx_psi
        name = 'Psi'
        names = model.names_psi
        if what != 'est':
            matrix = _set_values(model.parameters, matrix, what == 'start')
        ret[name] = pd.DataFrame(matrix, columns=names[1],
                                 index=names[0])
    if hasattr(model, 'mx_theta'):
        matrix = model.mx_theta
        name = 'Theta'
        names = model.names_theta
        if what != 'est':
            matrix = _set_values(model.parameters, matrix, what == 'start')
        ret[name] = pd.DataFrame(matrix, columns=names[1],
                                 index=names[0])

    if hasattr(model, 'mx_d'):
        matrix = model.mx_d
        name = 'D'
        names = model.names_d
        if what != 'est':
            matrix = _set_values(model.parameters, matrix, what == 'start')
        ret[name] = pd.DataFrame(matrix, columns=names[1],
                                 index=names[0])

    if hasattr(model, 'mx_v'):
        matrix = model.mx_v
        name = 'v'
        if what != 'est':
            matrix = _set_values(model.parameters, matrix, what == 'start')
        ret[name] = matrix[0, 0]

    if hasattr(model, 'mx_data_imp'):
        matrix = model.mx_data_imp
        name = 'Imputation'
        names = model.names_data_imp
        if what != 'est':
            matrix = _set_values(model.parameters, matrix, what == 'start')
        ret[name] = pd.DataFrame(matrix, columns=names[1],
                                 index=names[0])

    return ret


def inspect_list(model: Model, information='expected', std_est=False,
                 se_robust=False, index_names=False):
    """
    Get a pandas DataFrame containin a view of parameters estimates.

    Parameters
    ----------
    model : Model
        Model.
    information : str
        If 'expected', expected Fisher information is used. Otherwise,
        observed information is employed. If None, the no p-values are
        calculated. The default is 'expected'.
    std_est : bool
        If True, standardized coefficients are also returned as Std. Ests col.
        The default is False.
    se_robust : bool, optional
        If True, then robust SE are computed instead. Robustness here means
        that MLR-esque sandwich correction is applied. The default is False.
    index_names : bool, optional
        If True, then the returned DataFrame has parameter names as indices.

    Returns
    -------
    pd.DataFrame
        DataFrame with parameters information.

    """
    res = list()
    try:
        vals = model.param_vals
    except AttributeError:
        vals = ['Not estimated'] * len(model.parameters)
        information = None
        std_est = False
        se_robust = False
    if std_est:
        sigma, (_, c) = model.calc_sigma()
        w_cov = c @ model.mx_psi @ c.T
        stds = np.sqrt(w_cov.diagonal())
        std_full = np.sqrt(sigma.diagonal())
        if std_est == 'lv':
            std_full[:] = 1.0
    if information is not None:
        se = stats.calc_se(model, information=information, robust=se_robust)
        zscores = stats.calc_zvals(model, std_errors=se)
        pvals = stats.calc_pvals(model, z_scores=zscores)
    else:
        se = ['-'] * len(vals)
        zscores = pvals = se
    keys = list(model.parameters.keys())
    keys_active = [k for k in keys if model.parameters[k].active]
    param_names = list()
    # Beta
    if hasattr(model, 'mx_beta'):
        mx = model.mx_beta
        names = model.names_beta
        op = '~'
        for name, param in model.parameters.items():
            for loc in param.locations:
                if loc.matrix is mx:
                    param_names.append(name)
                    ind = loc.indices
                    a, b = names[0][ind[0]], names[1][ind[1]]
                    if param.active:
                        i = keys_active.index(name)
                        val = vals[i]
                        std = se[i]
                        zs = zscores[i]
                        pval = pvals[i]
                    else:
                        val = param.start
                        std = '-'
                        zs = '-'
                        pval = '-'
                    if std_est:
                        val_std = val / stds[ind[0]] * stds[ind[1]]
                        res.append((a, op, b, val, val_std, std, zs, pval))
                    else:
                        res.append((a, op, b, val, std, zs, pval))

    means = list()
    # Gamma1
    if hasattr(model, 'mx_gamma1'):
        mx = model.mx_gamma1
        names = model.names_gamma1
        op = '~'
        for name, param in model.parameters.items():
            for loc in param.locations:
                if loc.matrix is mx:
                    param_names.append(name)
                    ind = loc.indices
                    a, b = names[0][ind[0]], names[1][ind[1]]
                    if param.active:
                        i = keys_active.index(name)
                        val = vals[i]
                        std = se[i]
                        zs = zscores[i]
                        pval = pvals[i]
                    else:
                        val = param.start
                        std = '-'
                        zs = '-'
                        pval = '-'
                    if b == '1':
                        if std_est:
                            val_std = val * stds[ind[1]] / std_full[ind[0]]
                            means.append((a, op, b, val, val_std, std, zs, pval))
                        else:
                            means.append((a, op, b, val, std, zs, pval))
                    else:
                        if std_est:
                            val_std = val * stds[ind[1]] / std_full[ind[0]]
                            res.append((a, op, b, val, val_std, std, zs, pval))
                        else:
                            res.append((a, op, b, val, std, zs, pval))

    # Gamma2
    if hasattr(model, 'mx_gamma2'):
        mx = model.mx_gamma2
        names = model.names_gamma2
        op = '~'
        for name, param in model.parameters.items():
            for loc in param.locations:
                if loc.matrix is mx:
                    param_names.append(name)
                    ind = loc.indices
                    a, b = names[0][ind[0]], names[1][ind[1]]
                    if param.active:
                        i = keys_active.index(name)
                        val = vals[i]
                        std = se[i]
                        zs = zscores[i]
                        pval = pvals[i]
                    else:
                        val = param.start
                        std = '-'
                        zs = '-'
                        pval = '-'
                    if b == '1':
                        if std_est:
                            val_std = val / std_full[ind[0]]
                            means.append((a, op, b, val, val_std, std, zs, pval))
                        else:
                            means.append((a, op, b, val, std, zs, pval))
                    else:
                        if std_est:
                            val_std = val / std_full[ind[0]]
                            res.append((a, op, b, val, val_std, std, zs, pval))
                        else:
                            res.append((a, op, b, val, std, zs, pval))

    # Lambda
    if hasattr(model, 'mx_lambda'):
        mx = model.mx_lambda
        names = model.names_lambda
        op = '~'
        for name, param in model.parameters.items():
            for loc in param.locations:
                if loc.matrix is mx:
                    param_names.append(name)
                    ind = loc.indices
                    a, b = names[0][ind[0]], names[1][ind[1]]
                    if param.active:
                        i = keys_active.index(name)
                        val = vals[i]
                        std = se[i]
                        zs = zscores[i]
                        pval = pvals[i]
                    else:
                        val = param.start
                        std = '-'
                        zs = '-'
                        pval = '-'
                    if std_est:
                        val_std = val * stds[ind[1]]
                        if std_est != 'lv':
                            val_std /= std_full[ind[0]]
                        res.append((a, op, b, val, val_std, std, zs, pval))
                    else:
                        res.append((a, op, b, val, std, zs, pval))
    res.extend(means)
    means.clear()
    # Psi
    if hasattr(model, 'mx_psi'):
        mx = model.mx_psi
        names = model.names_psi
        op = '~~'
        obs_exo = set(model.vars['observed']) & model.vars['exogenous']
        for name, param in model.parameters.items():
            for loc in param.locations:
                if loc.matrix is mx:
                    ind = loc.indices
                    a, b = names[0][ind[0]], names[1][ind[1]]
                    if param.active:
                        i = keys_active.index(name)
                        val = vals[i]
                        std = se[i]
                        zs = zscores[i]
                        pval = pvals[i]
                    else:
                        if a in obs_exo and b in obs_exo:
                            continue
                        val = param.start
                        std = '-'
                        zs = '-'
                        pval = '-'
                    param_names.append(name)
                    if std_est:
                        val_std = val / stds[ind[0]] / stds[ind[1]]
                        res.append((a, op, b, val, val_std, std, zs, pval))
                    else:
                        res.append((a, op, b, val, std, zs, pval))
    # Theta
    if hasattr(model, 'mx_theta'):
        mx = model.mx_theta
        names = model.names_theta
        op = '~~'
        for name, param in model.parameters.items():
            for loc in param.locations:
                if loc.matrix is mx:
                    param_names.append(name)
                    ind = loc.indices
                    a, b = names[0][ind[0]], names[1][ind[1]]
                    if param.active:
                        i = keys_active.index(name)
                        val = vals[i]
                        std = se[i]
                        zs = zscores[i]
                        pval = pvals[i]
                    else:
                        val = param.start
                        std = '-'
                        zs = '-'
                        pval = '-'
                    if std_est:
                        val_std = val / std_full[ind[0]] / std_full[ind[1]]
                        res.append((a, op, b, val, val_std, std, zs, pval))
                    else:
                        res.append((a, op, b, val, std, zs, pval))

    # D -- Variance of random effects matrix
    if hasattr(model, 'mx_d'):
        mx = model.mx_d
        names = model.names_d
        op = 'RF'
        for name, param in model.parameters.items():
            for loc in param.locations:
                if loc.matrix is mx:
                    param_names.append(name)
                    ind = loc.indices
                    a, b = names[0][ind[0]], names[1][ind[1]]
                    if param.active:
                        i = keys_active.index(name)
                        val = vals[i]
                        std = se[i]
                        zs = zscores[i]
                        pval = pvals[i]
                    else:
                        val = param.start
                        std = '-'
                        zs = '-'
                        pval = '-'
                    if std_est:
                        val_est = val / std_full[ind[0]] / std_full[ind[1]]
                        res.append((a, op, b, val, val_est, std, zs, pval))
                    else:
                        res.append((a, op, b, val, std, zs, pval))
    # D_(i) -- Variance of random effects matrix in ModelGeneralizedEffects
    i = 1
    while hasattr(model, f'mx_d{i}'):
        mx = getattr(model, f'mx_d{i}')
        names = getattr(model, f'names_d{i}')
        op = f'RF{i}'
        for name, param in model.parameters.items():
            for loc in param.locations:
                if loc.matrix is mx:
                    param_names.append(name)
                    ind = loc.indices
                    a, b = names[0][ind[0]], names[1][ind[1]]
                    if param.active:
                        i = keys_active.index(name)
                        val = vals[i]
                        std = se[i]
                        zs = zscores[i]
                        pval = pvals[i]
                    else:
                        val = param.start
                        std = '-'
                        zs = '-'
                        pval = '-'
                    if std_est:
                        val_est = val / std_full[ind[0]] / std_full[ind[1]]
                        res.append((a, op, b, val, val_est, std, zs, pval))
                    else:
                        res.append((a, op, b, val, std, zs, pval))
        i += 1
    # v -- Variance of random effects variable
    if hasattr(model, 'mx_v'):
        mx = model.mx_v
        names = model.names_v
        op = 'RF(v)'
        for name, param in model.parameters.items():
            for loc in param.locations:
                if loc.matrix is mx:
                    param_names.append(name)
                    ind = loc.indices
                    a, b = names[0][ind[0]], names[1][ind[1]]
                    if param.active:
                        i = keys_active.index(name)
                        val = vals[i]
                        std = se[i]
                        zs = zscores[i]
                        pval = pvals[i]
                    else:
                        val = param.start
                        std = '-'
                        zs = '-'
                        pval = '-'
                    if std_est:
                        res.append((a, op, b, val, '-', std, zs, pval))
                    else:
                        res.append((a, op, b, val, std, zs, pval))
    # Data_imp -- Matrix of imputed data
    if hasattr(model, 'mx_data_imp'):
        mx = model.mx_data_imp
        names = model.names_data_imp
        op = '@'
        for name, param in model.parameters.items():
            for loc in param.locations:
                if loc.matrix is mx:
                    param_names.append(name)
                    ind = loc.indices
                    a, b = names[0][ind[0]], names[1][ind[1]]
                    if param.active:
                        i = keys_active.index(name)
                        val = vals[i]
                        std = se[i]
                        zs = zscores[i]
                        pval = pvals[i]
                    else:
                        val = param.start
                        std = '-'
                        zs = '-'
                        pval = '-'
                    res.append((a, op, b, val, std, zs, pval))
    if std_est:
        cols = ['lval', 'op', 'rval', 'Estimate', 'Est. Std', 'Std. Err',
                'z-value', 'p-value']
    else:
        cols = ['lval', 'op', 'rval', 'Estimate', 'Std. Err', 'z-value',
                'p-value']        
    res = pd.DataFrame(res, columns=cols)
    if index_names:
        res.index = param_names
    return res


def _set_values(params: dict, ref: np.ndarray, start=True):
    mx = ref.astype('O')
    for name, param in params.items():
        for loc in param.locations:
            if loc.matrix is ref:
                if start:
                    mx[loc.indices] = param.start
                else:
                    mx[loc.indices] = name
                if loc.symmetric:
                    mx[loc.indices[::-1]] = mx[loc.indices]
    return mx



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Contains Solver class that wraps around scipy and possibly other
optimization packages/procedures."""
from scipy.optimize import minimize, differential_evolution
from dataclasses import dataclass
from copy import copy
import logging
import numpy as np


@dataclass
class SolverResult:
    """Solver result."""

    fun: float
    success: bool
    n_it: int
    x: np.ndarray
    message: str
    name_method: str
    name_obj: str = ''

    def __str__(self):
        s = list()
        if self.name_obj:
            s.append('Name of objective: ' + self.name_obj)
        s.append('Optimization method: ' + self.name_method)  
        if self.success:
            s.append('Optimization successful.')
        s.append(str(self.message))
        s.append('Objective value: {:.3f}'.format(self.fun))
        s.append('Number of iterations: {}'.format(self.n_it))
        s.append('Params: {}'.format(' '.join('{:.3f}'.format(v)
                                              for v in self.x)))
        return '\n'.join(s)

    def __bool__(self):
        return self.success


class Solver():
    """Wraps around scipy. In future, custom optimization procedures might be
    present"""

    custom_methods = dict()

    def __init__(self, method: str, fun, grad, x0, bounds=None, constrs=None,
                 options=None, **kwargs):
        """
        Instantiate Solver.

        Parameters
        ----------
        method : str
            Method name.
        fun : TYPE
            Objective function f(x).
        grad : TYPE
            Gradient function grad(x). Can be None.
        bounds: list
            List of tuples specifying bound constraints on parameters.
        constrs: dict
            Dictionary in scipy format specifying constraints.
        options : dict, optional
            Dict of options to pass to optimizers. The default is
            {'maxiter': 1000}.
        **kwargs : dict
            Extra parameters.

        Returns
        -------
        None.

        """
        if options is None:
            options = dict()
        if 'maxiter' not in options:
            options['maxiter'] = 10000
        self.method = method
        self.fun = fun
        self.grad = grad
        self.options = options
        self.extra_options = kwargs
        self.start = x0
        self.bounds = bounds
        self.constraints = constrs if constrs else list()

    def solve(self):
        """
        Solve problem.

        Returns
        -------
        SolverResult
            Information on optimization result.

        """
        if self.method in self.custom_methods:
            return self.custom_methods[self.method]()
        res = self.scipy_solve()
        if not res.success:
            logging.warning('Solver didn''t converge, see SolverResult.')
        return res

    def scipy_solve(self):
        """
        Solve problem via scipy library.

        Returns
        -------
        SolverResult
            Information on optimization result.

        """
        if self.method == 'de':
            t = copy(self.extra_options)
            if 'b_max' in t:
                bmax = t['b_max']
                del t['b_max']
            else:
                bmax = 10
            b = self.bounds
            bs = list()
            for (a, b) in b:
                if a == None:
                    a = -bmax
                if b == None:
                    b = bmax
                bs.append((a, b))
            res = differential_evolution(self.fun, bounds=bs,
                       constraints=self.constraints, **t)
        else:
            res = minimize(self.fun, self.start, jac=self.grad, bounds=self.bounds,
                           constraints=self.constraints, options=self.options,
                           method=self.method, **self.extra_options)
        try:
            nit = res.nit
        except AttributeError:
            nit = np.nan
        return SolverResult(fun=res.fun, success=res.success, n_it=nit,
                            message=res.message, x=res.x,
                            name_method=self.method)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Multi-group analysis functionality.
"""

from . import Model
from .stats import calc_stats
from dataclasses import dataclass
import pandas as pd

@dataclass
class MultigroupResult:
    """Structure for printing info on multiple semopy runs, each per group."""

    groups: list
    estimates: dict
    stats: dict
    runs: dict
    n_obs: dict

    def __str__(self):
        s = 'semopy: Multi-Group analysis report\n\n'
        s += 'Groups: {}\n\n'.format(', '.join(self.groups))
        s += 'Number of observations per group:\n'
        s += '\n'.join('{}:\t{}'.format(g, self.n_obs[g]) for g in self.groups)
        s += '\n\nParameter estimates:\n'
        for i, g in enumerate(self.groups):
            s += f'Group {i + 1}: {g}\n'
            s += self.estimates[g].to_string() + '\n'
        return s
        

    


def multigroup(desc, data: pd.DataFrame, group: str, mod=Model,
               **kwargs):
    """
    Perform a multi-group analysis.

    Parameters
    ----------
    desc : str or dict
        Either text description of model or an unique model for each group.
    data : pd.DataFrame
        Dataset with a group column.
    group : str
        A name of the group columns.
    mod : method, optional
        Constructor of the model. The default is Model.
    **kwargs : dict
        Extra arguments that will be passed to the fit method.

    Returns
    -------
    res : MultigroupResult
        Printable MultigroupResult.

    """
    
    groups = data[group].unique()
    res = MultigroupResult(groups, dict(), dict(), dict(), dict())
    if type(desc) is str:
        desc = {g: desc for g in groups}
    for g in groups:
        m = mod(desc[g])
        g_data = data[data[group] == g].copy()
        res.n_obs[g] = len(g_data)
        res.runs[g] = m.fit(g_data, **kwargs)
        res.estimates[g] = m.inspect()
        res.stats[g] = calc_stats(m)
    return res



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""This module contains sympy-powered constraint parser."""
from sympy.parsing.sympy_parser import parse_expr
from sympy import lambdify, derive_by_array, Symbol
import numpy as np
import re

def parse_constraint(s: str, params: list):
    try:
        op = next(re.finditer('(?:>)|(?:<)|(?:=)',s))
        span = op.span()
        expr = s[:span[0]] + '-({})'.format(s[span[1]:])
        op = op.group()
        if op == '<':
            expr  = f'-({expr})'
    except StopIteration:
        raise SyntaxError('>,< or = must be present in constraint string.')
    if op in ('>', '<'):
        op = 'ineq'
    elif op == '=':
        op = 'eq'
    expr = expr.replace('^', '**')
    try:
        expr = parse_expr(expr)
    except SyntaxError:
        raise SyntaxError(f'Incorrect syntax for constraint {s}')
    params = list(map(Symbol, params))
    for v in expr.free_symbols:
        if v not in params:
            raise Exception(f"Can't add constraint: {v} is not a parameter.")
    grad = derive_by_array(expr, params)
    grad = list(map(lambda g: lambdify(params, g, 'numpy'), grad))
    grad_f = lambda x: np.array([g(*x) for g in grad])
    fun = lambdify(params, expr, 'numpy')
    f = lambda x: fun(*x)
    return {'type': op, 'fun': f, 'jac': grad_f}



# -*- coding: utf-8 -*-
"""Autoregressive effect model."""
from .effect_base import EffectBase
import numpy as np


class EffectAR(EffectBase):
    """
    AR(1) model.
    
    At the moment, only autoregressive model of the first order is supported.
    Feel free to introduce higher order models :)
    """
    def __init__(self, columns: str, dt=1, param=None, d_mode='diag'):
        """
        Instantiate EffectAR.

        Parameters
        ----------
        columns : str
            Name of column that corresponds to time at which the individual was
            observed. Should be numeric.
        dt : float, optional
            Min difference for time to be considered different. For example,
            if dt=1, |5 - 3| / dt = 2 - lag. The default is 1.
        param : float, optional
            If provided, then AR parameter is fixed to param. The default is
            None.
        d_mode : str
            Mode of D matrix. If "diag", then D has unique params on the
            diagonal. If "full", then D is fully parametrised. If
            "scale", then D is an identity matrix, multiplied by a single
            variance parameter (scalar). The default is "diag".

        Returns
        -------
        None.

        """
        
        super().__init__(columns, d_mode=d_mode)
        self.dt = dt
        self.param = param

    def load(self, i, model, data, clean_start=True, **kwargs):
        """
        Called by model new dataset is loaded.

        Parameters
        ----------
        order : int
            Identificator of effect in model. It is just an order of the effect
            among other effects as specified by user.
        model : ModelGeneralizedEffects
            Instance of ModelGeneralizedEffects that calls this method.
        data : pd.DataFrame
            Dataset that is being loaded. Should contain self.columns.
        clean_start : bool, optional
            If True, then parameters are (re)initialized. The model will use
            the ones already present in self.parameters vector otherwise. The
            default is True.

        Returns
        -------
        None.

        """
        super().load(i, model, data, clean_start, **kwargs)
        if clean_start:
            if self.param is not None:
                self.parameters = np.array([self.param])
            else:
                self.parameters = np.array([0.0])
        g = data[self.columns[0]].values
        triu = np.triu_indices(len(g))
        d = dict()
        for i in range(len(triu[0])):
            a = triu[0][i]
            b = triu[1][i]
            dt = abs(g[a] - g[b]) // self.dt
            lt = d.get(dt, None)
            if lt is None:
                lt = [list(), list()]
                d[dt] = lt
            lt[0].append(a)
            lt[1].append(b)
        for dt, (a, b) in d.items():
            d[dt] = [np.array(a), np.array(b)]
        self.d = d
        self.num_n = len(g)
        if self.param is not None:
            t = self.param
            self.param = None
            self.mx_k = self.calc_k(model)
            self.param = t

    def calc_k(self, model):
        if self.param is not None:
            return self.mx_k
        p = self.parameters[0]
        n = self.num_n
        mx = np.zeros((n, n))
        for dt, inds in self.d.items():
            rho = p ** dt
            mx[inds] = rho
            mx[inds[::-1]] = rho
        return mx

    def calc_k_grad(self, model):
        if self.param is not None:
            return list()
        n = self.num_n
        p = self.parameters[0]
        n = self.num_n
        mx = np.zeros((n, n))
        for dt, inds in self.d.items():
            if dt == 0:
                rho = 0
            else:
                rho = dt * (p ** (dt - 1))
            mx[inds] = rho
            mx[inds[::-1]] = rho
        return mx



# -*- coding: utf-8 -*-
"""Moving average effect model."""
from .effect_base import EffectBase
import numpy as np


class EffectMA(EffectBase):
    """
    MA(1) and MA(2) models.
    
    Number of new parameters introduced is equal to p in M(p).
    """
    def __init__(self, columns: str, order=1,
                 dt_bounds=None,
                 d_mode='diag'):
        """
        Instantiate EffectMA.

        Parameters
        ----------
        columns : str
            Name of column that corresponds to time at which the individual was
            observed. Should be numeric.
        order : int, optional
            Order "p" of the MA(p) model. The default is 1.
        dt_bounds : Tuple[tuple], optional
            List of tuples that of length p (order) that contain bounding boxes
            for autocorrelation function AF(dt=|t_i - t_j|). For instance,
            if p = 2 and dt_bounds = ((0, 1), (1, 4), (4,  10)), then AF for
            elements observed at times 121 and 118 is AF(1) as
            0 <= (121 - 118) = 3 < 4;  AF for elements observed at times 121
            and 300 is 0 as it is beyond any of the bounding boxes specified.
            If None, then ((0, 1), (1, 2), (2, 3)...(p, p + 1)) is assumed.
            The default is None.
        d_mode : str
            Mode of D matrix. If "diag", then D has unique params on the
            diagonal. If "full", then D is fully parametrised. If
            "scale", then D is an identity matrix, multiplied by a single
            variance parameter (scalar). The default is "diag".

        Returns
        -------
        None.

        """
        
        super().__init__(columns, d_mode=d_mode)
        self.ma_order = order
        if dt_bounds is None:
            self.dt_bounds = [(i, i + 1) for i in range(order + 1)]
        else:
            self.dt_bounds = dt_bounds

    def load(self, i, model, data, clean_start=True, **kwargs):
        """
        Called by model new dataset is loaded.

        Parameters
        ----------
        order : int
            Identificator of effect in model. It is just an order of the effect
            among other effects as specified by user.
        model : ModelGeneralizedEffects
            Instance of ModelGeneralizedEffects that calls this method.
        data : pd.DataFrame
            Dataset that is being loaded. Should contain self.columns.
        clean_start : bool, optional
            If True, then parameters are (re)initialized. The model will use
            the ones already present in self.parameters vector otherwise. The
            default is True.

        Returns
        -------
        None.

        """
        super().load(i, model, data, clean_start, **kwargs)
        if clean_start:
            self.parameters = np.array([0.0] * self.ma_order)
        g = data[self.columns[0]].values
        inds_a = [list() for _ in range(self.ma_order + 1)]
        inds_b = [list() for _ in range(self.ma_order + 1)]
        triu = np.triu_indices(len(g))
        for i in range(len(triu[0])):
            a = triu[0][i]
            b = triu[1][i]
            dt = abs(g[a] - g[b])
            for i, (c, d) in enumerate(self.dt_bounds):
                if c <= dt < d:
                    inds_a[i].append(a)
                    inds_b[i].append(b)
                    break
        self.inds = [(np.array(a), np.array(b)) for a, b in zip(inds_a,
                                                                inds_b)]
        self.num_n = len(g)


    def calc_k(self, model):
        params = self.parameters
        n = self.num_n
        mx = np.zeros((n, n))
        denom = 1 + (params ** 2).sum()
        for i, ind in enumerate(self.inds):
            if not len(ind[0]):
                continue
            if i == 0:
                num = 1
            else:
                num = params[i - 1]
            for p in range(1, self.ma_order - i + 1):
                num += params[p - 1] * params[p + i - 1]
            rho = num / denom
            mx[ind] = rho
            mx[ind[::-1]] = rho
        return mx
            

    def calc_k_grad(self, model):
        params = self.parameters
        n = self.num_n
        m = len(params)
        grad = [np.zeros((n, n)) for _ in range(m)]
        mx = np.zeros((n, n))
        denom = 1 + (params ** 2).sum()
        for i, ind in enumerate(self.inds):
            dro = np.array([0.0] * m)
            if i == 0:
                num = 1
            else:
                num = params[i - 1]
                dro[i - 1] += 1
            for p in range(1, self.ma_order - i + 1):
                num += params[p - 1] * params[p + i - 1]
                dro[p - 1] += params[p + i - 1] 
                dro[p + i - 1] += params[p - 1] 
            dro /= denom
            d1 = - 2 * num / (denom ** 2)
            for p in range(m):
                dro[p] += d1 * params[p]
                mx = grad[p]
                mx[ind] += dro[p]
                mx[ind[::-1]] += dro[p]
        return grad



# -*- coding: utf-8 -*-
"""The module is responsible for generating model descriptions/specifications
given some parameters."""
from copy import deepcopy
import random


def generate_cfa(n_lat: int, n_inds: int, p_join=0.0, max_join=1,
                 base_lat='eta', base_ind='y'):
    """
    Generate random CFA model.

    Parameters
    ----------
    n_lat : int, tuple
        Number of latent factors in the model. If tuple, then the number of
        factors is chosen randomly from the interval.
    n_inds : int, tuple
        Number of indicators per factor in the model. If tuple, then the number
        of indicators is chosen randomly from the interval for each factor.
    p_join : float
        A probability that the indicator will be shared with another latent
        factor. Closely related to max_join, as the test whether to add the
        indicator to another random variable is evaluated for each other latent
        factor until either there is no factors left or until it has beed added
        to max_join factors. The default is 0.0.
    max_join : int
        Maximal number of factors per indicator.
    base_lat : str, tuple
        Base of a name for latent variable. If tuple, then the tuple should
        be of length n_lat and then it contains names of the variables. The
        default is 'eta'.
    base_ind : str
        Base of a name for indocator variable. The default is 'y'.

    Returns
    -------
    Adjacency dictionary: factor -> indicators.

    """
    res = dict()
    if type(n_lat) in (tuple, list):
        n_lat = random.randint(n_lat[0], n_lat[1])        
    ind_count = 0
    for i in range(n_lat):
        m = n_inds
        if type(m) in (tuple, list):
            m = random.randint(m[0], m[1])
        if type(base_lat) in (tuple, list):
            lat = base_lat[i]
        else:
            lat = f'{base_lat}{i+1}'
        lt = list()
        for _ in range(m):
            ind_count += 1
            ind = f'{base_ind}{ind_count}'
            lt.append(ind)
        res[lat] = lt
    tres = deepcopy(res)
    for lat, inds in tres.items():
        for a in inds:
            for latb in tres:
                if latb != lat and random.uniform(0, 1) < p_join:
                    res[latb].append(a)
    return res

def generate_sem(n_endo: int, n_exo: int, n_cycles: int, p_edge=0.3, cfa=None,
                 strict_exo=True, base_endo='x', base_exo='g'):
    """
    Generate SEM model (CFA + PA).

    Parameters
    ----------
    n_endo : int, tuple
        Number of endogenous variables. If tuple, then the number of
        endogenous variables is chosen randomly from the interval.
    n_exo : int, tuple
        Number of exogenous variables (or integer range). If 0, then actual
        number of exogenous variable will still be 1, but it is in fact will
        be one of the "endogenous" (not anymore!) variables.
    n_cycles : int, tuple
        Number of cycles. If tuple, then the number of
        cycles is chosen randomly from the interval.
    p_edge: float
        Probability that an extra edge (i.e. the one that is not required
        to make graph connected) will be added. The default is 0.3.
    cfa : dict, optional
        CFA part of the model. The default is None.
    strict_exo : bool, optional
        If True, then there will be exactly n_exo exogenous variables. The
        default is True.
    base_endo : str, tuple, optional
        Base of a name for endogenous variable. If tuple, then the tuple should
        be of length n_exo and then it contains names of the variables. The
        default is 'x'.
    base_exo : str, tuple, optional
        Base of a name for exogenous variable. If tuple, then the
        tuple should be of length n_exo and then it contains names of
        the variables. The default is 'g'.

    Returns
    -------
    Adjacency dictionary: rval -> lval.
    (i.e. lval ~ rval)
    """
    if type(n_endo) in (tuple, list):
        n_endo = random.randint(n_endo[0], n_endo[1])
    if type(n_cycles) in (tuple, list):
        n_cycles = random.randint(n_cycles[0], n_cycles[1])
    if type(n_exo) in (tuple, list):
        n_exo = random.randint(n_exo[0], n_exo[1])
    if type(base_exo) in (tuple, list) and len(base_exo) == n_exo:
        exos = base_exo
    else:
        exos = [f'{base_exo}{i + 1}' for i in range(n_exo)]
    if type(base_endo) is str:
        endos = [f'{base_endo}{i + 1}' for i in range(n_endo)]
    else:
        endos = base_endo
    if cfa:
        endos += list(cfa.keys())
        res = deepcopy(cfa)
    else:
        res = dict()
    random.shuffle(endos)
    nodes = exos + endos
    for i in range(len(nodes) - 1):
        sx = max(len(exos), i + 1)
        rv = nodes[i]
        lv = random.choice(nodes[sx:])
        lt = res.get(rv, list())
        lt.append(lv)
        for i in range(sx, len(nodes)):
            if nodes[i] != lv and random.uniform(0, 1) < p_edge:
                lt.append(nodes[i])
        res[rv] = lt
    
    if strict_exo:
        endogenous = set()
        for _, ins in res.items():
            endogenous.update(ins)
        left_out = set(endos) - endogenous
        for v in left_out:
            i = nodes.index(v)
            if i:
                rv = random.choice(nodes[:i])
                if rv not in res:
                    res[rv] = list()
                res[rv].append(v)
    if n_cycles:
        n = len(exos)
        s_nodes = nodes[n:]
        random.shuffle(s_nodes)
        for v in s_nodes:
            i = nodes.index(v)
            if i > n:
                v2 = random.choice(nodes[n:i])
                if v not in res:
                    res[v] = [v2]
                else:
                    res[v].append(v2)
                n_cycles -= 1
                if not n_cycles:
                    break
    return res


def dict_to_desc(d: dict, lats=None):
    """
    Translate dictionary to string that can be used by semopy models.

    Parameters
    ----------
    d : dict
        Dict containing mapping rval->lval (i.e. lval ~ rval).
    lats : dict, optional
        Measurement part. The default is None.

    Returns
    -------
    String that can be fed to semopy model.

    """
    measurement_part = list()
    structural_part = list()
    defines = set()
    d = deepcopy(d)
    to_rem = set()
    if lats:
        for lat, inds in lats.items():
            if lat not in d:
                defines.add(lat)   
            else:
                inds = list(filter(lambda x: x not in d, inds))
                rvals = ' + '.join(inds)
                s = f'{lat} =~ {rvals}'
                measurement_part.append(s)
                lt = d.get(lat, None)
                if lt:
                    for ind in inds:
                        lt.remove(ind)
                    if not lt:
                        to_rem.add(lat)
    for lat in to_rem:
        del d[lat]
    mappings = dict()
    for rval, lvals in d.items():
        for lval in lvals:
            if lval in mappings:
                lt = mappings[lval]
            else:
                lt = list()
            lt.append(rval)
            mappings[lval] = lt
    for lval, rvals in mappings.items():
        rvals = ' + '.join(rvals)
        s = f'{lval} ~ {rvals}'
        structural_part.append(s)
    res = str()
    if measurement_part:
        res += '# Measurement part:\n' + '\n'.join(measurement_part) + '\n'
    if defines:
        res += '## Latent variables that ought to be defined:\n'
        s = ' '.join(defines)
        res += f'DEFINE(latent) {s}\n'
    if structural_part:
        res += '# Structural part:\n' + '\n'.join(structural_part)
    return res


def generate_desc(n_endo: int, n_exo: int, n_lat: int, n_inds: int,
                  n_cycles: int, p_join=0.0, max_join=1, p_edge=0.3,
                  strict_exo=True, base_lat='eta', base_ind='y', base_endo='x',
                  base_exo='g',):
    """
    Generate string description of a SEM model, suitable for semopy models.

    Parameters
    ----------
    n_lat : int, tuple
        Number of latent factors in the model. If tuple, then the number of
        factors is chosen randomly from the interval.
    n_inds : int, tuple
        Number of indicators per factor in the model. If tuple, then the number
        of indicators is chosen randomly from the interval for each factor.
    p_join : float
        A probability that the indicator will be shared with another latent
        factor. Closely related to max_join, as the test whether to add the
        indicator to another random variable is evaluated for each other latent
        factor until either there is no factors left or until it has beed added
        to max_join factors. The default is 0.0.
    max_join : int
        Maximal number of factors per indicator.
    base_lat : str, tuple
        Base of a name for latent variable. If tuple, then the tuple should
        be of length n_lat and then it contains names of the variables. The
        default is 'eta'.
    base_ind : str
        Base of a name for indocator variable. The default is 'y'.

    Returns
    -------
    String description of model, suitable for semopy models.
    """
    cfa = generate_cfa(n_lat, n_inds, p_join=p_join, max_join=max_join,
                       base_lat=base_lat, base_ind=base_ind,)
    sem = generate_sem(n_endo, n_exo, n_cycles, p_edge=p_edge, cfa=cfa,
                       strict_exo=strict_exo, base_endo=base_endo,
                       base_exo=base_exo)
    return dict_to_desc(sem, cfa)



# -*- coding: utf-8 -*-
"""Effect with a pre-defied K kernel similiarity matrix."""
from sklearn.gaussian_process.kernels import Kernel
from .effect_base import EffectBase
from copy import deepcopy
import numpy as np


class EffectKernel(EffectBase):
    """
    Kernel similiarity effect.
    
    If active parameter is False, then this effect introduces no extra
    parameters other than those that come along the D matrix. Kernels are
    provided by sklearn.gaussian_process.kernels submodule.
    """
    def __init__(self, columns: str, kernel: Kernel, params: dict,
                 active=False, d_mode='diag'):
        """
        Instantiate EffectMatern.

        Parameters
        ----------
        columns : str
            Name of column that corresponds to individuals group id. Should
            match the appropriate row/column in the K dataframe.
        kernel : sklearn.gaussian_process.kernels.Kernel
            Kernel that is used to compute pairwise similarity matrix.
        params : dict
            A dictionary of parameters that are passed to kernel.
        active : bool, optional
            If True, then kernel parameters are active. Otherwise, they are
            fixed to those provided by params. The default is False.
        d_mode : str
            Mode of D matrix. If "diag", then D has unique params on the
            diagonal. If "full", then D is fully parametrised. If
            "scale", then D is an identity matrix, multiplied by a single
            variance parameter (scalar). The default is "diag".

        Returns
        -------
        None.

        """
        
        super().__init__(columns, d_mode=d_mode)
        self.active = active
        if active:
            params = deepcopy(params)
            k = kernel(**params)
            ps = k.get_params()
            t = list()
            for h in k.hyperparameters:
                if h.name not in params:
                    params[h.name] = ps[h.name]
                t.append((h.name, h.n_elements))
            self.param_names = t
        self.kernel = kernel
        self.kernel_params = params

    def load(self, i, model, data, clean_start=True, **kwargs):
        """
        Called by model new dataset is loaded.
        
        Here, Effects are configured from the data. self.parameters must be
        initialised after invoking this method.
        Parameters
        ----------
        order : int
            Identificator of effect in model. It is just an order of the effect
            among other effects as specified by user.
        model : ModelGeneralizedEffects
            Instance of ModelGeneralizedEffects that calls this method.
        data : pd.DataFrame
            Dataset that is being loaded. Should contain self.columns.
        clean_start : bool, optional
            If True, then parameters are (re)initialized. The model will use
            the ones already present in self.parameters vector otherwise. The
            default is True.

        Returns
        -------
        None.

        """
        super().load(i, model, data, clean_start, **kwargs)
        if clean_start:
            if self.active:
                p = self.kernel_params
                params = list()
                for n, m in self.param_names:
                    ps = p[n]
                    if m != 1:
                        params.extend(ps)
                    else:
                        params.append(ps)
                self.parameters = np.array(params)
            else:
                self.parameters = np.array([])
        if not self.active:
            k = self.kernel(**self.kernel_params)(data[self.columns].values)
            self.mx_k = k
        else:
            self.mx_c = data[self.columns].values    

    def calc_k(self, model):
        if not self.active:
            return self.mx_k
        else:
            d = self.kernel_params
            i = 0
            p = self.parameters
            for n, m in self.param_names:
                if m == 1:
                    d[n] = p[i]
                else:
                    d[n] = p[i:m]
                i += m
            return self.kernel(**d)(self.mx_c)
    
    def calc_k_grad(self, model):
        if not self.active:
            return list()
        else:
            d = self.kernel_params
            i = 0
            p = self.parameters
            for n, m in self.param_names:
                if m == 1:
                    d[n] = p[i]
                else:
                    d[n] = p[i:m]
                i += m
            k = self.kernel(**d)(self.mx_c, eval_gradient=True)[1]
            return [k[:, :, i] for i in range(len(self.parameters))]

    def get_bounds(self):
        k = self.kernel(**self.kernel_params)
        bounds = list()
        for h in k.hyperparameters:
            for a, b in h.bounds:
                bounds.append((a, b))
        return bounds


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Graphviz wrapper to visualie SEM models."""
from .model import Model
import logging
try:
    import graphviz
    __GRAPHVIZ = True
except ModuleNotFoundError:
    logging.info("No graphviz package found, visualization method is "
                 "unavailable")
    __GRAPHVIZ = False
    

def semplot(mod: Model, filename: str, inspection=None, plot_covs=False,
            plot_exos=True, images=None, engine='dot', latshape='circle',
            plot_ests=True, std_ests=False, show=False):
    """
    Draw a SEM diagram.

    Parameters
    ----------
    mod : Model
        Model instance.
    filename : str
        Name of file where to plot is saved.
    inspection : pd.DataFrame, optional
        Parameter estimates as returned by Model.inspect(). The default is
        None.
    plot_covs : bool, optional
        If True, covariances are also drawn. The default is False.
    plot_exos: bool, optional
        If False, exogenous variables are not plotted. It might be useful,
        for example, in GWAS setting, where a number of exogenous variables,
        i.e. genetic markers, is oblivious. Has effect only with ModelMeans or
        ModelEffects. The default is True.
    images : dict, optional
        Node labels can be replaced with images. It will be the case if a map
        variable_name->path_to_image is provided. The default is None.
    engine : str, optional
        Graphviz engine name to use. The default is 'dot'.
    latshape : str, optional
        Graphviz-compaitable shape for latent variables. The default is
        'circle'.
    plot_ests : bool, optional
        If True, then estimates are also plotted on the graph. The default is
        True.
    std_ests : bool, optional
        If True and plot_ests is True, then standardized values are plotted
        instead. The default is False.
    show : bool, optional
        If True, the 

    Returns
    -------
    Graphviz graph.

    """
    if not __GRAPHVIZ:
        raise ModuleNotFoundError("No graphviz module is installed.")
    if type(mod) is str:
        mod = Model(mod)
    if not hasattr(mod, 'last_result'):
        plot_ests = False
    if inspection is None:
        inspection = mod.inspect(std_est=std_ests)
    if images is None:
        images = dict()
    if std_ests:
        inspection['Estimate'] = inspection['Est. Std']
    t = filename.split('.')
    filename, ext = '.'.join(t[:-1]), t[-1]
    g = graphviz.Digraph('G', format=ext, engine=engine)
    
    g.attr(overlap='scale', splines='true')
    g.attr('edge', fontsize='12')
    g.attr('node', shape=latshape, fillcolor='#cae6df', style='filled')
    for lat in mod.vars['latent']:
        if lat in images:
            g.node(lat, label='', image=images[lat])
        else:
            g.node(lat, label=lat)
    
    g.attr('node', shape='box', style='')
    for obs in mod.vars['observed']:
        if obs in images:
            g.node(obs, label='', image=images[obs])
        else:
            g.node(obs, label=obs)

    regr = inspection[inspection['op'] == '~']
    all_vars = mod.vars['all']
    try:
        exo_vars = mod.vars['observed_exogenous']
    except KeyError:
        exo_vars = set()
    for _, row in regr.iterrows():
        lval, rval, est = row['lval'], row['rval'], row['Estimate']
        if (rval not in all_vars) or (~plot_exos and rval in exo_vars) or\
            (rval == '1'):
            continue
        if plot_ests:
            pval = row['p-value']
            label = '{:.3f}'.format(float(est))
            if pval !='-':
                label += r'\np-val: {:.2f}'.format(float(pval))
        else:
            label = str()
        g.edge(rval, lval, label=label)
    if plot_covs:
        covs = inspection[inspection['op'] == '~~']
        for _, row in covs.iterrows():
            lval, rval, est = row['lval'], row['rval'], row['Estimate']
            if lval == rval:
                continue
            if plot_ests:
                pval = row['p-value']
                label = '{:.3f}'.format(float(est))
                if pval !='-':
                    label += r'\np-val: {:.2f}'.format(float(pval))
            else:
                label = str()
            g.edge(rval, lval, label=label, dir='both', style='dashed')
    g.render(filename, view=show)
    return g


# -*- coding: utf-8 -*-
"""
Holzinger-Swineford 1939 dataset and CFA model.

The classic Holzinger and Swineford (1939) dataset consists of mental ability
test scores of seventh- and eighth-grade children from two different schools
(Pasteur and Grant-White). 
"""
import pandas as pd
import os

__desc = '''visual =~ x1 + x2 + x3
textual =~ x4 + x5 + x6
speed =~ x7 + x8 + x9'''



__filename = '%s/holzinger_swineford39_data.csv' % \
    os.path.dirname(os.path.abspath(__file__))


def get_model():
    """
    Retrieve model description in semopy syntax.

    Returnsunivariate_regression
    -------
    str
        Model's description.

    """
    return __desc


def get_data():
    """
    Retrieve dataset.

    Returns
    -------
    pd.DataFrame
        Dataset.

    """
    return pd.read_csv(__filename, index_col=0, sep=',')


# -*- coding: utf-8 -*-
"""The module is responsible for generating data for a model.
"""
from .. import ModelMeans
import pandas as pd
import numpy as np
import scipy


def data_exogenous(shape: tuple):
    return np.random.normal(size=shape)


def generate_data(model: ModelMeans, n: int, drop_lats=True,
                  generator_exo=data_exogenous):
    """
    Generate data for a given model with parameters fit.

    Parameters
    ----------
    model : ModelMeans
        ModelMeans that will generate data. It is assumed that the model used
        is just an auxiliarily structure that is returned by other generator
        functions, yet it is fine to use with any custom ModelMeans.
    n : int
        Number of data samples.
    drop_lats : bool, optional
        If True, latent variables are dropped from the dataset. The default is
        True.

    Returns
    -------
    Padnas DataFrame with the generated data.
    """
    if hasattr(model, 'calc_t'):
        sigma, (m, c) = model.calc_sigma()
        mean = model.calc_mean(m)
        t = model.calc_t(sigma)
        l = model.calc_l(sigma)
        tr_l = np.trace(l)
        res = scipy.stats.matrix_normal.rvs(mean, l / tr_l, t)
        res = np.append(res, model.mx_g, axis=0)
        cols = model.vars['observed'] + model.vars['observed_exogenous']
    else:
        c = np.linalg.inv(np.identity(model.mx_beta.shape[0]) - model.mx_beta)
        if c.shape[0]:
            epsilon = np.random.multivariate_normal(mean=np.zeros(c.shape[0]),
                                                    cov=model.mx_psi, size=n)
        else:
            epsilon = None
        lambc = model.mx_lambda @ c
        m = model.mx_lambda.shape[0]
        delta = np.random.multivariate_normal(mean=np.zeros(m),
                                              cov=model.mx_theta, size=n)
        try:
            exo = model.vars['observed_exogenous']
            if generator_exo:
                gamma = generator_exo((len(exo), n))
                try:
                    i = exo.index('1')
                    gamma[i, :] = 1
                except ValueError:
                    pass
            else:
                gamma = model.mx_g
            res = lambc @  model.mx_gamma1 @ gamma + delta.T
            if epsilon is not None:
                res += lambc @ epsilon.T
            res += model.mx_gamma2 @ gamma
            res = np.append(res, gamma, axis=0)
            cols = model.vars['observed'] + exo
        except KeyError:
            res = lambc @ epsilon.T + delta.T
            cols = model.vars['observed']
    df = pd.DataFrame(res.T, columns=cols)
    if not drop_lats and not hasattr(model, 'calc_t') and epsilon is not None:
        omega = c @ epsilon.T
        for i, v in enumerate(model.vars['inner']):
            if v in model.vars['latent']:
                df[v] = omega[i, :]
    if '1' in df.columns:
        df = df.drop('1', axis=1)
    return df



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""This module implements polychoric and polyserial correlations."""
from statsmodels.stats.correlation_tools import corr_nearest
from scipy.optimize import minimize, minimize_scalar
from itertools import chain, product, combinations
from scipy.stats import norm, mvn
from .utils import cor
import pandas as pd
import numpy as np


def bivariate_cdf(lower, upper, corr, means=[0, 0], var=[1, 1]):
    """
    Estimates an integral of bivariate pdf.

    Estimates an integral of bivariate pdf given integration lower and 
    upper limits. Consider using relatively big (i.e. 20 if using default mean
    and variance) lower and/or upper bounds when integrating to/from infinity.
    Parameters
    ----------
    lower : float
        Lower integration bounds.
    upper : float
        Upper integration bounds.
    corr : float
        Correlation coefficient between variables.
    means : list, optional
        Mean values of variables. The default is [0, 0].
    var : list, optional
        Variances of variables. The default is [1, 1].

    Returns
    -------
    float
        P(lower[0] < x < upper[0], lower[1] < y < upper[1]).

    """
    s = np.array([[var[0], corr], [corr, var[1]]])
    return mvn.mvnun(lower, upper, means, s)[0]

def univariate_cdf(lower, upper, mean=0, var=1):
    """
    Estimate an integral of univariate pdf.

    Estimate an integral of univariate pdf given integration lower and 
    upper limits. Consider using relatively big (i.e. 20 if using default mean
    and variance) lower and/or upper bounds when integrating to/from infinity.
    Parameters
    ----------
    lower : float
        Lower integration bound..
    upper : float
        Upper integration bound..
    mean : float, optional
        Mean value of the variable. The default is 0.
    var : float, optional
        Variance of the variable. The default is 1.

    Returns
    -------
    float
        P(lower < x < upper).

    """
    return mvn.mvnun([lower], [upper], [mean], [var])[0]

def estimate_intervals(x, inf=10):
    """
    Estimate intervals of the polytomized underlying latent variable.

    Parameters
    ----------
    x : np.ndarray
        An array of values the ordinal variable..
    inf : float, optional
        A numerical infinity substitute. The default is 10.

    Returns
    -------
    np.ndarray
        An array containing polytomy intervals.
    np.ndarray
        An array containing indices of intervals corresponding to each entry.

    """
    x_f = x[~np.isnan(x)]
    u, counts = np.unique(x_f, return_counts=True)
    sz = len(x_f)
    cumcounts = np.cumsum(counts[:-1])
    u = [np.where(u == sample)[0][0] + 1 for sample in x]
    return list(chain([-inf], (norm.ppf(n / sz) for n in cumcounts), [inf])), u

def polyserial_corr(x, y, x_mean=None, x_var=None, x_z=None, x_pdfs=None,
                    y_ints=None, scalar=True):
    """
    Estimate polyserial correlation.
    
    Estimate polyserial correlation between continious variable x and
    ordinal variable y.
    Parameters
    ----------
    x : np.ndarray
        Data sample corresponding to x.
    y : np.ndarray
        Data sample corresponding to ordinal variable y.
    x_mean : float, optional
        Mean of x (calculate if not provided). The default is None.
    x_var : float, optional
        Variance of x (calculate if not provided). The default is None.
    x_z : np.ndarray, optional
        Stabdardized x (calculated if not provided). The default is None.
    x_pdfs : np.ndarray, optional
        x's logpdf sampled at each point (calculated if not provided). The
        default is None.
    y_ints : list, optional
        Polytomic intervals of an underlying latent variable
        correspoding to y (calculated if not provided) as returned by
        estimate_intervals.. The default is None.
    scalar : bool, optional
        If true minimize_scalar is used instead of SLSQP.. The default is True.

    Returns
    -------
    float
        A polyserial correlation coefficient for x and y..

    """
    if x_mean is None:
        x_mean = np.nanmean(x)
    if x_var is None:
        x_var = np.nanvar(x)
    if y_ints is None:
        y_ints = estimate_intervals(y)
    if x_z is None:
        x_z = (x - x_mean) / x_var
    if x_pdfs is None:
        x_pdfs = norm.logpdf(x, x_mean, x_var)
    ints, inds = y_ints
    def transform_tau(tau, rho, z):
        return (tau - rho * z) / np.sqrt(1 - rho ** 2)
    def sub_pr(k, rho, z):
        i = transform_tau(ints[k], rho, z)
        j = transform_tau(ints[k - 1], rho, z)
        return univariate_cdf(j, i)
    def calc_likelihood(rho):
        return -sum(pdf + np.log(sub_pr(ind, rho, z))
                    for z, ind, pdf in zip(x_z, inds, x_pdfs))
    def calc_likelihood_derivative(rho):
        def sub(k, z):
            i = transform_tau(ints[k], rho, z)
            j = transform_tau(ints[k - 1], rho, z)
            a = norm.pdf(i) * (ints[k] * rho - z)
            b = norm.pdf(j) * (ints[k - 1] * rho - z)
            return a - b
        t = (1 - rho ** 2) ** 1.5
        return -sum(sub(ind, z) / sub_pr(ind, rho, z)
                   for x, z, ind in zip(x, x_z, inds) if not np.isnan(x)) / t
    if not scalar:
        res = minimize(calc_likelihood, [0.0], jac=calc_likelihood_derivative,
                       method='SLSQP', bounds=[(-1.0, 1.0)]).x[0]
    else:
        res = minimize_scalar(calc_likelihood, bounds=(-1, 1),
                              method='bounded').x
    return res

def polychoric_corr(x, y, x_ints=None, y_ints=None):
    """
    Estimate polyserial correlation between ordinal variables x and y.

    Parameters
    ----------
    x : np.ndarray
        Data sample corresponding to x.
    y : np.ndarray
        Data sample corresponding to y.
    x_ints : list, optional
        Polytomic intervals of an underlying latent variable correspoding to y
        (calculated if not provided) as returned by estimate_intervals. The
        default is None.
    y_ints : list, optional
        Polytomic intervals of an underlying latent variable correspoding to y
        (calculated if not provided) as returned by estimate_intervals. The
        default is None.

    Returns
    -------
    float
        A polychoric correlation coefficient for x and y.

    """
    if x_ints is None:
        x_ints = estimate_intervals(x)
    if y_ints is None:
        y_ints = estimate_intervals(y)
    x_ints, x_inds = x_ints
    y_ints, y_inds = y_ints
    p, m = len(x_ints) - 1, len(y_ints) - 1
    n = np.zeros((p, m))
    for a, b in zip(x_inds, y_inds):
        if not (np.isnan(a) or np.isnan(b)):
            n[a - 1, b - 1] += 1
    def calc_likelihood(r):
        return -sum(np.log(bivariate_cdf([x_ints[i], y_ints[j]],
                                 [x_ints[i + 1], y_ints[j + 1]], r)) * n[i, j]
                    for i in range(p) for j in range(m))
    return minimize_scalar(calc_likelihood, bounds=(-1, 1), method='bounded').x
                

def hetcor(data, ords=None, nearest=False):
    """
    Compute a heterogenous correlation matrix.

    Parameters
    ----------
    data : pd.DataFrame or np.ndarray
        DESCRIPTION.
    ords : list, optional
        Names of ordinal variables if data is DataFrame or indices of
        ordinal numbers if data is np.array. If ords are None then ordinal
        variables will be determined automatically. The default is None.
    nearest : bool, optional
        If True, then nearest PD correlation matrix is returned instead. The
        default is False.

    Returns
    -------
    cor : pd.DataFrame
        A heterogenous correlation matrix.

    """
    if type(data) is np.ndarray:
        cov = cor(data)
        if ords is None:
            ords = set()
            for i in range(data.shape[1]):
                if len(np.unique(data[:, i])) / data.shape[0] < 0.3:
                    ords.add(i)
        conts = set(range(data.shape[1])) - set(ords)
    else:
        cov = data.corr()
        if ords is None:
            ords = set()
            for var in data:
                if len(data[var].unique()) / len(data[var]) < 0.3:
                    ords.add(var)
        conts = set(data.columns) - set(ords)
    c_means = {v: np.nanmean(data[v]) for v in conts}
    c_vars = {v: np.nanvar(data[v]) for v in conts}
    c_z = {v: (data[v] - c_means[v]) / c_vars[v] for v in conts}
    c_pdfs = {v: norm.logpdf(data[v], c_means[v], c_vars[v]) for v in conts}
    o_ints = {v: estimate_intervals(data[v]) for v in ords}

    for c, o in product(conts, ords):
        cov[c][o] = polyserial_corr(data[c], data[o], x_mean=c_means[c],
                                    x_var=c_vars[c], x_z=c_z[c],
                                    x_pdfs=c_pdfs[c], y_ints=o_ints[o])
        cov[o][c] = cov[c][o]
    for a, b in combinations(ords, 2):
        cov[a][b] = polychoric_corr(data[a], data[b], o_ints[a], o_ints[b])
        cov[b][a] = cov[a][b]
    if nearest:
        if type(cov) is pd.DataFrame:
            names = cov.columns
            cov = corr_nearest(cov,threshold=0.05)
            cov = pd.DataFrame(cov, columns=names, index=names)
        else:
            cov = corr_nearest(cov, threshold=0.05)
    return cov



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Random Effects SEM."""
import pandas as pd
import numpy as np
from .model_means import ModelMeans
from .utils import chol_inv, chol_inv2, cov, calc_zkz
from .univariate_blup import blup
from scipy.linalg import solve_sylvester
from collections import defaultdict
from .solver import Solver
from itertools import combinations


class ModelEffects(ModelMeans):
    """
    Random Effects model.

    Random Effects SEM can be interpreted as a generalization of Linear Mixed
    Models (LMM) to SEM.
    """

    matrices_names = tuple(list(ModelMeans.matrices_names) + ['d'])
    symb_rf_covariance = '~R~'

    def __init__(self, description: str, mimic_lavaan=False, baseline=False,
                 cov_diag=False, intercepts=True, d_mode='diag', effects=None):
        """
        Instantiate Random Effects SEM.

        Parameters
        ----------
        description : str
            Model description in semopy syntax.
        mimic_lavaan: bool, optional
            If True, output variables are correlated and not conceptually
            identical to indicators. lavaan treats them that way, but it's
            less computationally effective. The default is False.
        baseline : bool, optional
            If True, the model will be set to baseline model.
            Baseline model here is an independence model where all variables
            are considered to be independent with zero covariance. Only
            variances are estimated. The default is False.
        cov_diag : bool, optional
            If cov_diag is True, then there are no covariances parametrised
            unless explicitly specified. The default is False.
        intercepts: bool, optional
            If True, intercepts are also modeled. Intercept terms can be
            accessed via "1" symbol in a regression equation, i.e. x1 ~ 1. The
            default is False.
        d_mode : str, optional
            Mode of D matrix. If "diag", then D has unique params on the
            diagonal. If "full", then D is fully parametrised. If
            "scale", then D is an identity matrix, multiplied by a single
            variance parameter (scalar). Utilised only if effect names
            are not provided. The default is "diag".
        effects : list, optional
            List of effects name. Must correspond to cetrain columns in data.
            If None and effects are not provided in syntax, then matrix D
            is parametrised in accordance to d_mode. The default is None.
        Returns
        -------
        None.

        """
        self.dict_effects[self.symb_rf_covariance] = self.effect_rf_covariance
        self.d_mode = d_mode
        if effects is None:
            effects = set()
        if len(effects) > 1:
            raise Exception("ModelEffects supports only one random effect."
                            " Consider using ModelGeneralizedEffects.")
        self.effects_names = effects
        self.effects_loadings = defaultdict(float)
        super().__init__(description, mimic_lavaan=mimic_lavaan, 
                         cov_diag=cov_diag, baseline=baseline,
                         intercepts=intercepts)

        self.objectives = {'REML': (self.obj_reml, self.grad_reml),
                           'REML2': (self.obj_reml2, self.grad_reml2),
                           'ML': (self.obj_matnorm, self.grad_matnorm)}

    def before_classification(self, effects: dict, operations: dict):
        """
        Preprocess effects and operations if necessary before classification.

        Parameters
        ----------
        effects : dict
            Dict returned from parse_desc.

        operations: dict
            Dict of operations as returned from parse_desc.

        Returns
        -------
        None.

        """
        super().before_classification(effects, operations)
        symb = self.symb_rf_covariance
        regr = self.symb_regression
        eff_names = self.effects_names
        loadings = self.effects_loadings
        eff_regr = effects[regr]
        eff_rf = effects[symb]
        for v, rvs in eff_regr.items():
            to_rem = list()
            for rv in rvs:
                if rv in eff_names:
                    eff_rf[v][v] = None
                    loadings[v] = 0.05
                    to_rem.append(rv)
                    continue
            for rv in to_rem:
                del rvs[rv]

    def preprocess_effects(self, effects: dict):
        """
        Run a routine just before effects are applied.

        Used to apply random effect variance
        Parameters
        -------
        effects : dict
            Mapping opcode->lvalues->rvalues->multiplicator.

        Returns
        -------
        None.

        """
        super().preprocess_effects(effects)
        mode = self.d_mode
        symb = self.symb_rf_covariance
        regr = self.symb_regression
        obs = self.vars['observed']
        eff_names = self.effects_names
        loadings = self.effects_loadings
        if not eff_names:
            if mode in ('diag', 'full'):
                for v in obs:
                    t = effects[symb][v]
                    if v not in t:
                        t[v] = None
                        loadings[v] = list()
                if mode == 'full':
                    for a, b in combinations(obs, 2):
                        t = effects[symb][a]
                        tt = effects[symb][b]
                        if (v not in t) and (v not in tt):
                            t[b] = None
            else:
                if mode != 'scale':
                    raise Exception(f'Unknown mode "{mode}".')
                param = 'paramD'
                for v in obs:
                    t = effects[symb][v][v] = param
                    loadings[v] = list()

    def build_d(self):
        """
        D matrix is a covariance matrix for random effects across columns.

        Returns
        -------
        np.ndarray
            Matrix.
        tuple
            Tuple of rownames and colnames.

        """
        names = self.vars['observed']
        n = len(names)
        mx = np.zeros((n, n))
        return mx, (names, names)

    def load(self, data, group=None, k=None, cov=None, clean_slate=False,
             n_samples=None, obj='ML'):
        """
        Load dataset.

        Parameters
        ----------
        data : pd.DataFrame
            Data with columns as variables.
        group : str, optional
            Name of column with group labels. If not provided, predefined
            columns are used.
        k : pd.DataFrame, optional
            Covariance matrix across rows, i.e. kinship matrix. If None,
            identity is assumed. The default is None.
        cov : pd.DataFrame, optional
            Pre-computed covariance/correlation matrix. Used only for variance
            starting values. The default is None.
        clean_slate : bool, optional
            If True, resets parameters vector. The default is False.
        n_samples : int, optional
            Redunant for ModelEffects. The default is None.
        obj : str, optional
            Objective fuction name necessary to do the initial data
            preparation. The default is 'ML'.

        KeyError
            Rises when there are missing variables from the data.
        Exception
            Rises when group parameter is None.
        Returns
        -------
        None.

        """
        if data is None:
            if not hasattr(self, 'mx_data'):
                raise Exception("Data must be provided.")
            if clean_slate:
                self.prepare_params()
            return
        else:
            data = data.copy()
        obs = self.vars['observed']
        exo = self.vars['observed_exogenous']
        if self.intercepts:
            data['1'] = 1.0
        cols = data.columns
        missing = (set(obs) | set(exo)) - set(set(cols))
        if missing:
            t = ', '.join(missing)
            raise KeyError('Variables {} are missing from data.'.format(t))
        self.load_data(data, k=k, covariance=cov, group=group)
        if obj == 'REML':
            if self.__loaded != 'REML':
                self.load_reml()
        elif obj == 'ML':
            if self.__loaded != 'ML':
                self.load_ml()
        self.load_starting_values()
        if clean_slate or not hasattr(self, 'param_vals'):
            self.prepare_params()

    def _fit(self, obj='REML', solver='SLSQP', **kwargs):
        fun, grad = self.get_objective(obj)
        solver = Solver(solver, fun, grad, self.param_vals,
                        constrs=self.constraints,
                        bounds=self.get_bounds(),
                        **kwargs)
        res = solver.solve()
        res.name_obj = obj
        self.param_vals = res.x
        self.update_matrices(res.x)
        self.last_result = res
        return res

    def fit(self, data=None, group=None, k=None, cov=None, obj='ML',
            solver='SLSQP', clean_slate=False, regularization=None, **kwargs):
        """
        Fit model to data.

        Parameters
        ----------
        data : pd.DataFrame, optional
            Data with columns as variables. The default is None.
        group : str, optioal
            Name of column in data with group labels. Overrides effect names
            provided by the constructor. The default is None.
        cov : pd.DataFrame, optional
            Pre-computed covariance/correlation matrix. The default is None.
        obj : str, optional
            Objective function to minimize. Possible values are 'REML', 'ML'.
            The default is 'REML'.
        solver : TYPE, optional
            Optimizaiton method. Currently scipy-only methods are available.
            The default is 'SLSQP'.
        clean_slate : bool, optional
            If False, successive fits will be performed with previous results
            as starting values. If True, parameter vector is reset each time
            prior to optimization. The default is False.
        regularization
            Special structure as returend by create_regularization function.
            If not None, then a regularization will be applied to a certain
            parameters in the model. The default is None.

        Raises
        ------
        Exception
            Rises when attempting to use MatNorm in absence of full data.

        Returns
        -------
        SolverResult
            Information on optimization process.

        """
        self.load(data=data, cov=cov, group=group, k=k,
                  clean_slate=clean_slate, obj=obj)
        if not hasattr(self, 'mx_data'):
            raise Exception('Full data must be supplied.')
        if obj == 'REML':
            # if self.__loaded != 'REML':
            #     self.load_reml()
            res_reml = self._fit(obj='REML', solver=solver, **kwargs)
            self.load_ml(fake=True)
            sigma, (self.mx_m, _) = self.calc_sigma()
            self.mx_r_inv = chol_inv(self.calc_l(sigma))
            self.mx_w_inv = self.calc_t_inv(sigma)[0]
            res_reml2 = self._fit(obj='REML2', solver=solver, **kwargs)
            return (res_reml, res_reml2)
        elif obj == 'ML':
            # if self.__loaded != 'ML':
            #     self.load_ml()
            res = self._fit(obj='ML', solver=solver, **kwargs)
            return res
        else:
            raise NotImplementedError(f'Unknown objective {obj}.')      

    def effect_rf_covariance(self, items: dict):
        """
        Work through random effects covariance operation.

        Parameters
        ----------
        items : dict
            Mapping lvalues->rvalues->multiplicator.

        Returns
        -------
        None.

        """
        mx = self.mx_d
        rows, cols = self.names_d
        for lv, rvs in items.items():
            for rv, mult in rvs.items():
                name = None
                try:
                    val = float(mult)
                    active = False
                except (TypeError, ValueError):
                    active = True
                    if mult is not None:
                        if mult != self.symb_starting_values:
                            name = mult
                        else:
                            active = False
                    val = None
                if name is None:
                    self.n_param_cov += 1
                    name = '_c%s' % self.n_param_cov
                i, j = rows.index(lv), cols.index(rv)
                ind = (i, j)
                if i == j:
                    bound = (0, None)
                    symm = False
                else:
                    if self.baseline:
                        continue
                    bound = (None, None)
                    symm = True
                self.add_param(name, matrix=mx, indices=ind, start=val,
                               active=active, symmetric=symm, bound=bound)

    def set_fim_means(self):
        """
        Substitute true FIM matrix with means-only FIM matrix.

        A trick to reduce GWAS time as only mean components are subject to
        analysis.
        Returns
        -------
        None.

        """
        
        self.calc_fim = self.calc_fim_means

    def load_starting_values(self):
        """
        Load starting values for parameters from empirical data.

        Returns
        -------
        None.

        """
        trans_data = self.mx_data_transformed.copy()
        obs = self.vars['observed']
        loadings = self.effects_loadings
        for v in loadings:
            i = obs.index(v)
            y = trans_data[i]
            k = self.mx_s
            up, s = blup(y - y.mean(), k)
            if s[0] > 0.01 and s[1] > 0.01:
                trans_data[i] -= up
                loadings[v] = s[1]
        cov = self.mx_cov.copy()
        self.mx_cov = np.cov(trans_data)
        if len(self.mx_cov.shape) < 2:
            self.mx_cov = self.mx_covcov.reshape((1, 1))
        super().load_starting_values()
        self.mx_cov = cov


    '''
    ----------------------------LINEAR ALGEBRA PART---------------------------
    ----------------------The code below is responsible-----------------------
    ------------------for covariance structure computations-------------------
    '''

    '''
    ---------------------------R and W matrices-------------------------------
    '''

    def calc_l(self, sigma: np.ndarray):
        """
        Calculate covariance across columns matrix R.

        Parameters
        ----------
        sigma : np.ndarray
            Sigma matrix.

        Returns
        -------
        tuple
            R matrix.

        """
        n = self.num_n
        return n * sigma + self.mx_d * self.trace_zkz

    def calc_l_grad(self, sigma_grad: list):
        """
        Calculate gradient of R matrix.

        Parameters
        ----------
        sigma_grad : list
            Sigma gradient values.

        Returns
        -------
        grad : list
            Gradient of R matrix.

        """
        grad = list()
        n = self.num_n
        for g, df in zip(sigma_grad, self.mx_diffs):
            g = n * g
            if df[6] is not None:  # D
                g += df[6] * self.trace_zkz
            grad.append(g)
        return grad

    def calc_t_inv(self, sigma: np.ndarray):
        """
        Calculate inverse and logdet of covariance across rows matrix W.

        This function estimates only inverse of W. There was no need in package
        to estimate W.
        Parameters
        ----------
        sigma : np.ndarray
            Sigma matrix.

        Returns
        -------
        tuple
        R^{-1} and ln|R|.

        """
        w = self.calc_t(sigma)
        if np.any(w < 1e-9):
            raise np.linalg.LinAlgError
        return 1 / w, np.sum(np.log(w))

    def calc_t(self, sigma: np.ndarray):
        """
        Calculate W matrix.

        Parameters
        ----------
        sigma : np.ndarray
            Sigma matrix.

        Returns
        -------
        np.ndarray
            W matrix.

        """
        tr_sigma = np.trace(sigma)
        tr_d = np.trace(self.mx_d)
        return tr_d * self.mx_s + tr_sigma

    def calc_t_grad(self, sigma_grad: list):
        """
        Calculate gradient of W matrix.

        Parameters
        ----------
        sigma_grad : list
            Gradient of Sigma matrix.

        Returns
        -------
        grad : list
            Gradient of W.

        """
        grad = list()
        for g, df in zip(sigma_grad, self.mx_diffs):
            if len(g.shape):
                g = np.trace(g) * self.mx_i_n
            if df[6] is not None:  # D
                g += np.trace(df[6]) * self.mx_s
            grad.append(g)
        return grad

    def calc_t_inv_grad(self, inv_w: np.ndarray, sigma_grad: list):
        """
        Calculate gradient of W inverse and logdet matrix.

        Parameters
        ----------
        inv_w : np.ndarray
            Inverse of W matrix.
        sigma_grad : list
            Gradient of Sigma matrix.

        Returns
        -------
        grad : list
            Gradient of inverse of W.
        grad_logdet : list
            Gradient of logdet of W.

        """
        grad, grad_logdet = list(), list()
        inv_w_t = inv_w * self.mx_s
        inv_w_d = -(inv_w ** 2)
        inv_w_d_t = inv_w_d * self.mx_s
        for g, df in zip(sigma_grad, self.mx_diffs):
            gw = g
            gl = g
            if len(g.shape):
                tr = np.trace(g)
                gw = tr * inv_w_d
                gl = tr * inv_w
            if df[6] is not None:  # D
                tr = np.trace(df[6])
                gw += tr * inv_w_d_t
                gl += tr * inv_w_t
            grad.append(gw)
            grad_logdet.append(np.sum(gl))
        return grad, grad_logdet

    '''
    ---------------------Preparing structures for a more-----------------------
    ------------------------efficient computations-----------------------------
    '''

    def load_data(self, data: pd.DataFrame, group=None, k=None,
                  covariance=None):
        """
        Load dataset from data matrix.

        Parameters
        ----------
        data : pd.DataFrame
            Dataset with columns as variables and rows as observations.
        group : str, optional
            Name of column that correspond to group labels. If not provided,
            items from effects_names are used. The default is None.
        k : pd.DataFrame or tuple
            Covariance matrix betwen groups. If None, then it's assumed to be
            an identity matrix. Alternatively, a tuple of (ZKZ^T, S, Q) can be
            provided where ZKZ^T = Q S Q^T an eigendecomposition of ZKZ^T. S
            must be provided in the vector/list form. The default is None.
        covariance : pd.DataFrame, optional 
            Custom covariance matrix. The default is None.

        Returns
        -------
        None.

        """
        obs = self.vars['observed']
        if type(k) in (tuple, list):
            if len(k) != 3:
                raise Exception("Both ZKZ^T and its eigendecomposition must "
                                "be provided.")
        if group is None:
            group = next(iter(self.effects_names))
        self.mx_g_orig = data[self.vars['observed_exogenous']].values.T
        if len(self.mx_g_orig.shape) != 2:
            self.mx_g_orig = self.mx_g_orig[np.newaxis, :]
        self.mx_g = self.mx_g_orig
        self.mx_data = data[obs].values
        self.n_samples, self.n_obs = self.mx_data.shape
        self.num_m = len(set(self.vars['observed']) - self.vars['latent'])
        self.passed_k = k
        if type(k) is tuple:
            self.mx_zkz, self.mx_sk, self.mx_q = k
            self._ktuple = True
        else:
            self._ktuple = False
            self.mx_zkz = calc_zkz(data[group], k)
        self.__loaded = None
        self.load_cov(covariance[obs].loc[obs]
                      if covariance is not None else cov(self.mx_data))

    def load_ml(self, fake=False):
        self.trace_zkz = np.trace(self.mx_zkz)
        if self._ktuple:
            self.mx_s = self.mx_sk
            q = self.mx_q
        else:
            s, q = np.linalg.eigh(self.mx_zkz)
            self.mx_s, self.mx_q = s, q
        self.mx_data_transformed = self.mx_data.T @ q
        self.mx_g = self.mx_g_orig @ q
        self.num_n = self.mx_data_transformed.shape[1]
        self.mx_i_n = np.ones(self.num_n)
        self.__loaded = 'ML'

    def load_reml(self):
        g = self.mx_g_orig
        try:
            s = np.identity(g.shape[1]) - g.T @ chol_inv(g @ g.T) @ g
        except ValueError:
            raise Exception("REML should not be used when there are no"
                            " either intercepts or exogenous variables in "
                            "Gamma matrices.")
        d, q = np.linalg.eigh(s)
        rank_dec = 0
        for i in d:
            if abs(i) < 1e-8:
                rank_dec += 1
            else:
                break
        d = np.diag(d)[rank_dec:, :]
        a = d @ q.T
        azkza = a @ self.mx_zkz @ a.T
        self.trace_zkz = np.trace(azkza)
        s, q = np.linalg.eigh(azkza)
        self.mx_s = s
        self.mx_data_transformed = self.mx_data.T @ a.T @ q
        self.num_n = self.mx_data_transformed.shape[1]
        self.mx_i_n = np.ones(self.num_n)
        self.__loaded = 'REML'

    '''
    ---------------Matrix Variate Normal Restricted Maximum Likelihood---------
    '''

    def obj_reml(self, x: np.ndarray):
        """
        Restricted loglikelihood of matrix-variate normal distribution.

        Parameters
        ----------
        x : np.ndarray
            Parameters vector.

        Returns
        -------
        float
            Loglikelihood.

        """
        self.update_matrices(x)
        sigma, _ = self.calc_sigma()
        try:
            r = self.calc_l(sigma)
            r_inv, logdet_r = chol_inv2(r)
            w_inv, logdet_w = self.calc_t_inv(sigma)
        except np.linalg.LinAlgError:
            return np.inf
        mx = self.mx_data_transformed
        tr_r = np.trace(r)
        n, m = self.num_n, self.num_m
        r_center = r_inv @ mx
        center_w = mx * w_inv
        tr = tr_r * np.einsum('ji,ji->', center_w, r_center)
        return tr + m * logdet_w + n * logdet_r - n * m * np.log(tr_r)

    def grad_reml(self, x: np.ndarray):
        """
        Gradient of REML objective of matrix-variate normal distribution.

        Parameters
        ----------
        x : np.ndarray
            Parameters vector.

        Returns
        -------
        np.ndarray
            Gradient of REML objective.

        """
        self.update_matrices(x)
        grad = np.zeros_like(x)
        sigma, (m, c) = self.calc_sigma()
        try:
            r = self.calc_l(sigma)
            r_inv = chol_inv(r)
            w_inv, _ = self.calc_t_inv(sigma)
        except np.linalg.LinAlgError:
            grad[:] = np.inf
            return grad
        center = self.mx_data_transformed
        A = r_inv @ center
        B = (center * w_inv).T
        tr_ab = np.einsum('ij,ji->', A, B)
        tr_r = np.trace(r)
        V1 = center.T @ A
        V3 = B @ r_inv
        V2 = self.num_n * r_inv / tr_r - A @ V3
        sigma_grad = self.calc_sigma_grad(m, c)
        r_grad = self.calc_l_grad(sigma_grad)
        w_grad, w_grad_logdet = self.calc_t_inv_grad(w_inv, sigma_grad)
        n, m = self.num_n, self.num_m
        for i, (d_r, d_w, d_l) in enumerate(zip(r_grad, w_grad,
                                                w_grad_logdet)):
            g = 0.0
            tr_long = 0.0
            if len(d_r.shape):
                tr_long += np.einsum('ij,ji->', V2, d_r)
                tr_dr = np.trace(d_r)
                g += tr_dr * tr_ab
                g -= m * n * tr_dr / tr_r
            if len(d_w.shape):
                tr_long += np.einsum('ii,i->', V1, d_w)
                g += m * d_l
            g += tr_r * tr_long
            grad[i] = g
        return grad

    '''
    ------------------Matrix Variate REML (II-nd stage)-----------------------
    '''

    def obj_reml2(self, x: np.ndarray):
        """
        Loglikelihood of matrix-variate normal distribution given Sigma.

        For a second stage of REML estimation.
        Parameters
        ----------
        x : np.ndarray
            Parameters vector.

        Returns
        -------
        float
            Loglikelihood.

        """
        self.update_matrices(x)
        mean = self.calc_mean(self.mx_m)
        center = self.mx_data_transformed - mean
        r_center = self.mx_r_inv @ center
        center_w = center * self.mx_w_inv
        return np.einsum('ji,ji->', center_w, r_center)

    def grad_reml2(self, x: np.ndarray):
        """
        Gradient of loglikelihood of matrix-variate normal distribution.

        For a second stage of REML estimation.
        Parameters
        ----------
        x : np.ndarray
            Parameters vector.

        Returns
        -------
        np.ndarray
            Gradient of MatNorm objective.

        """
        self.update_matrices(x)
        grad = np.zeros_like(x)
        center = self.mx_data_transformed - self.calc_mean(self.mx_m)
        t =  (center * self.mx_w_inv).T @ self.mx_r_inv
        mean_grad = self.calc_mean_grad_reml()
        for i, g in enumerate(mean_grad):
            if len(g.shape):
                grad[i] = -2 * np.einsum('ij,ji->', g, t)
        return grad

    '''
    ------------------Matrix Variate Normal Maximum Likelihood-----------------
    '''

    def obj_matnorm(self, x: np.ndarray):
        """
        Loglikelihood of matrix-variate normal distribution.

        Parameters
        ----------
        x : np.ndarray
            Parameters vector.

        Returns
        -------
        float
            Loglikelihood.

        """
        self.update_matrices(x)
        sigma, (m, _) = self.calc_sigma()
        try:
            r = self.calc_l(sigma)
            r_inv, logdet_r = chol_inv2(r)
            w_inv, logdet_w = self.calc_t_inv(sigma)
        except np.linalg.LinAlgError:
            return np.inf
        mean = self.calc_mean(m)
        center = self.mx_data_transformed - mean
        tr_r = np.trace(r)
        m, n = self.num_m, self.num_n
        r_center = r_inv @ center
        center_w = center * w_inv
        tr = tr_r * np.einsum('ij,ij->', center_w, r_center)
        return tr + m * logdet_w + n * logdet_r - n * m * np.log(tr_r)

    def grad_matnorm(self, x: np.ndarray):
        """
        Gradient of loglikelihood of matrix-variate normal distribution.

        Parameters
        ----------
        x : np.ndarray
            Parameters vector.

        Returns
        -------
        np.ndarray
            Gradient of MatNorm objective.

        """
        self.update_matrices(x)
        grad = np.zeros_like(x)
        sigma, (m, c) = self.calc_sigma()
        try:
            r = self.calc_l(sigma)
            r_inv = chol_inv(r)
            w_inv, _ = self.calc_t_inv(sigma)
        except np.linalg.LinAlgError:
            grad[:] = np.inf
            return grad
        mean = self.calc_mean(m)
        center = self.mx_data_transformed - mean
        A = r_inv @ center
        B = (center * w_inv).T
        tr_ab = np.einsum('ij,ji->', A, B)
        tr_r = np.trace(r)
        V1 = center.T @ A
        V3 = B @ r_inv
        V2 = self.num_n * r_inv / tr_r - A @ V3

        sigma_grad = self.calc_sigma_grad(m, c)
        mean_grad = self.calc_mean_grad(m, c)
        r_grad = self.calc_l_grad(sigma_grad)
        w_grad, w_grad_logdet = self.calc_t_inv_grad(w_inv, sigma_grad)
        n, m = self.num_n, self.num_m
        for i, (d_m, d_r, d_w, d_l) in enumerate(zip(mean_grad, r_grad,
                                                     w_grad, w_grad_logdet)):
            g = 0.0
            tr_long = 0.0
            if len(d_m.shape):
                tr_long -= 2 * np.einsum('ij,ji->', V3, d_m)
            if len(d_r.shape):
                tr_long += np.einsum('ij,ji->', V2, d_r)
                tr_dr = np.trace(d_r)
                g += tr_dr * tr_ab
                g -= m * n * tr_dr / tr_r
            if len(d_w.shape):
                tr_long += np.einsum('ii,i->', V1, d_w)
                g += m * d_l
            g += tr_r * tr_long
            grad[i] = g
        return grad

    '''
    -----------------------Fisher Information Matrix---------------------------
    '''

    def calc_fim_ml(self, inverse=False):
        """
        Calculate Fisher Information Matrix.

        Exponential-family distributions are assumed.
        Parameters
        ----------
        inverse : bool, optional
            If True, function also returns inverse of FIM. The default is
            False.
        reml : bool, optional
            If True, then adjustment for REML is made. The default is False.

        Returns
        -------
        np.ndarray
            FIM.
        np.ndarray, optional
            FIM^{-1}.

        """
        sigma, (m, c) = self.calc_sigma()
        sigma_grad = self.calc_sigma_grad(m, c)
        mean_grad = self.calc_mean_grad(m, c)
        t_inv = self.calc_t_inv(sigma)[0]
        l = self.calc_l(sigma)
        try:
            l_inv = chol_inv(l)
        except np.linalg.LinAlgError:
            l_inv = np.linalg.pinv(l)
        l_grad = self.calc_l_grad(sigma_grad)
        t_grad = self.calc_t_grad(sigma_grad)
        tr_l = np.trace(l)
        a = [t_inv * g if not np.isscalar(g) else None for g in t_grad]
        b = [l_inv @ g if not np.isscalar(g) else None for g in l_grad]
        tr_a = [np.sum(g) if g is not None else g for g in a]
        tr_b = [np.trace(g) if g is not None else g for g in b]
        m_t = [g * t_inv if not np.isscalar(g) else None for g in mean_grad]
        m_l = [g.T @ l_inv if not np.isscalar(g) else None for g in mean_grad]
        al = [np.trace(g) / tr_l if not np.isscalar(g) else None
              for g in l_grad]
        param_len = len(self.param_vals)
        fim = np.zeros((param_len, param_len))
        n = self.num_n
        m = self.num_m
        n, m = m, n
        for i in range(param_len):
            for j in range(i, param_len):
                mean = 0
                cov = 0
                mtj = m_t[j]
                mli = m_l[i]
                ai = a[i]
                aj = a[j]
                bi = b[i]
                bj = b[j]
                trai = tr_a[i]
                traj = tr_a[j]
                trbi = tr_b[i]
                trbj = tr_b[j]
                alphai = al[i]
                alphaj = al[j]
                if mli is not None and mtj is not None:
                    mean += tr_l * np.einsum('ij,ji', mli, mtj)
                if ai is not None and aj is not None:
                    cov += n * (ai * aj).sum()
                    cov += m * np.einsum('ij,ji', bi, bj)
                    cov += trai * trbj + trbi * traj
                    cov += n * m * alphai * alphaj
                    cov -= n * alphaj * trai + m * alphaj * trbi
                    cov -= n * alphai * traj + m * alphai * trbj
                fim[i, j] = mean + cov / 2
                fim[j, i] = fim[i, j]
        if inverse:
            fim_inv = np.linalg.pinv(fim)
            return (fim, fim_inv)
        return fim

    def calc_fim_means(self, inverse=False):
        """
        Calculate Fisher Information Matrix for mean components only.

        Exponential-family distributions are assumed. Useful to fascilate GWAS
        as we usually don't care about other parameters.
        Parameters
        ----------
        inverse : bool, optional
            If True, function also returns inverse of FIM. The default is
            False.

        Returns
        -------
        np.ndarray
            FIM.
        np.ndarray, optional
            FIM^{-1}.

        """
        sigma, (m, c) = self.calc_sigma()
        mean_grad = self.calc_mean_grad(m, c)
        t_inv = self.calc_t_inv(sigma)[0]
        l = self.calc_l(sigma)
        try:
            l_inv = chol_inv(l)
        except np.linalg.LinAlgError:
            l_inv = np.linalg.pinv(l)
        tr_l = np.trace(l)
        m_t = [g * t_inv if not np.isscalar(g) else None for g in mean_grad]
        m_l = [g.T @ l_inv if not np.isscalar(g) else None for g in mean_grad]
        inds = list()
        c = 0
        for g in m_t:
            if g is not None:
                inds.append(c)
                c += 1
            else:
                inds.append(None)
        fim_means = np.zeros((len(inds), len(inds)))
        param_len = len(self.param_vals)
        fim = np.zeros((param_len, param_len))
        for i in range(param_len):
            for j in range(i, param_len):
                mtj = m_t[j]
                mli = m_l[i]
                if mli is not None and mtj is not None:
                    fim[i, j] = tr_l * np.einsum('ij,ji', mli, mtj)
                    fim[j, i] = fim[i, j]
                    it, jt = inds[i], inds[j]
                    fim_means[it, jt] = fim[i, j]
                    fim_means[jt, it] = fim[i, j]
        if inverse:
            fim_inv = np.linalg.pinv(fim)
            return (fim, fim_inv)
        return fim

    '''
    -------------------------Prediction method--------------------------------
    '''
    
    def predict(self, x: pd.DataFrame, k=None, group=None):
        """
        Predict data given certain observations.
        
        Uses conditional expectation of the normal distribution method.

        Parameters
        ----------
        x : pd.DataFrame
            DataFrame with missing variables either not present at all, or
            with missing entries set to NaN.
        k : pd.DataFrame, optional
            K relatedness matrix for elements in x. If x is None, then
            observations are assumed to be independent. The default is None.
        group : str, optional
            Columns name where group ID is stored. The default is None.

        Returns
        -------
        None.

        """
        sigma, (m, _) = self.calc_sigma()
        obs = self.vars['observed']
        exo = self.vars['observed_exogenous']
        result = x.copy()
        for v in obs:
            if v not in result:
                result[v] = np.nan
        for v in exo:
            if v not in result:
                if v == '1':
                    result[v] = 1
                else:
                    result[v] = 0
        result = result[obs + exo]
        old_g = self.mx_g.copy()
        self.mx_g = result[exo].values.T
        mean = self.calc_mean(m).reshape((-1, 1), order='F')
        self.mx_g = old_g
        k = calc_zkz(x[group], k)
        l = sigma * len(x) + self.mx_d * np.trace(k)
        t = np.trace(sigma) * np.identity(len(x)) + k * np.trace(self.mx_d)
        l /= np.trace(l)
        cov = np.kron(t, l)
        data = result[obs].values.T
        data_shape = data.shape
        data = data.reshape((-1, 1), order='F')
        missing = np.isnan(data).flatten()
        present = ~missing
        cov12 = cov[missing][:, present]
        cov22 = np.linalg.pinv(cov[present][:, present])
        mean_m = mean[missing]
        mean_p = mean[present]
        preds = mean_m
        if len(present):
            preds += cov12 @ cov22 @ (data[present] - mean_p)
        data[missing] = preds
        data = data.reshape(data_shape, order='F').T
        result = pd.DataFrame(data, columns=obs)
        return result

    def predict_factors(self, x: pd.DataFrame):
        """
        Fast factor estimation method via MAP. Requires complete data.

        Parameters
        ----------
        x : pd.DataFrame
            Complete data of observed variables.

        Returns
        -------
        Factor scores.

        """
        lats = self.vars['latent']
        num_lat = len(lats)
        if num_lat == 0:
            return pd.DataFrame([])
        inners = self.vars['inner']
        obs = self.vars['observed']
        obs_exo = self.vars['observed_exogenous']
        g = []
        for v in obs_exo:
            if v == '1':
                g.append([1] * len(x))
            else:
                g.append(x[v])
        g = np.array(g)
        x = x[obs].values.T
        lambda_h = self.mx_lambda[:, :num_lat]
        lambda_x = self.mx_lambda[:, num_lat:]
        c = np.linalg.inv(np.identity(self.mx_beta.shape[0]) - self.mx_beta)
        c_1 = c[:num_lat, :]
        c_2 = c[num_lat:, :]
        g1 = self.mx_gamma1; g2 = self.mx_gamma2;
        M_h = x - (g2 + lambda_x @ c_2 @ g1) @ g
        t = lambda_x @ c_2
        L_zh = (t @ self.mx_psi @ t.T + self.mx_theta) * (x.shape[1])
        tr_sigma = np.trace(L_zh) / x.shape[1]
        L_zh += self.mx_d * np.trace(self.mx_zkz)
        tr_lzh = np.trace(L_zh)
        try:
            L_zh = chol_inv(L_zh)
        except np.linalg.LinAlgError:
            L_zh = np.linalg.pinv(L_zh)
        T_zh = np.identity(x.shape[1]) * tr_sigma
        T_zh += self.mx_zkz * np.trace(self.mx_d)
        try:
            T_zh = chol_inv(T_zh)
        except np.linalg.LinAlgError:
            T_zh = np.linalg.pinv(T_zh)
        t = lambda_h.T @ L_zh
        A = tr_lzh * t @ M_h @ T_zh
        A_0 = tr_lzh * t @ lambda_h
        try:
            L_h = chol_inv(c_1 @ self.mx_psi @ c_1.T)
        except np.linalg.LinAlgError:
            L_h = np.linalg.pinv(c_1 @ self.mx_psi @ c_1.T)
        M = c_1 @ g1 @ g
        A_1 = L_h @ M
        try:
            inv_A0 = chol_inv(A_0)
        except np.linalg.LinAlgError:
            inv_A0 = np.linalg.pinv(A_0)
        A_2 = inv_A0 @ L_h
        A_hat = inv_A0 @ (A + A_1)
        H = solve_sylvester(A_2, T_zh, A_hat)
        return pd.DataFrame(H.T, columns=filter(lambda v: v in lats, inners))


# -*- coding: utf-8 -*-
"""Effect with a K matern covariance/similiarity matrix."""
from sklearn.gaussian_process.kernels import Matern
from .effect_kernel import EffectKernel


class EffectMatern(EffectKernel):
    """
    Matern covariance matrices are useful in spatial analysis.
    
    If 'active' parameter is False, then this effect introduces no extra
    parameters other than those that come along the D matrix.
    """
    def __init__(self, columns: str, nu=float('inf'), rho=1.0, 
                 active=False, d_mode='diag'):
        """
        Instantiate EffectMatern.

        Parameters
        ----------
        columns : str
            Name of column that corresponds to individuals group id. Should
            match the appropriate row/column in the K dataframe.
        nu : float, optional
            Num parameter of the matern kernel. The default is inf.
        rho : float, optional
            Rho parameter of the matern kernel, i.e. length. The default is 1.
        active : bool, optional
            If True, then rho is an active parameter that is optimized.
            Otherwise, it is fixed. The default is False.
        d_mode : str
            Mode of D matrix. If "diag", then D has unique params on the
            diagonal. If "full", then D is fully parametrised. If
            "scale", then D is an identity matrix, multiplied by a single
            variance parameter (scalar). The default is "diag".

        Returns
        -------
        None.

        """
        p = {'length_scale': rho,
             'nu': nu}
        super().__init__(columns, kernel=Matern, params=p, active=active,
                         d_mode=d_mode)



# -*- coding: utf-8 -*-
"""This is do-nothing effect that servers entirely for demonstrational and
testing purposes"""
from .effect_base import EffectBase
import numpy as np

class EffectBlank(EffectBase):
    """
    This effect does nothing. K matrix is just a static identity matrix.
    
    The only purposes of this effect are testing and providing guidelines for
    possible developers.
    """
    def __init__(self, d_mode='identity'):
        super().__init__(None, d_mode=d_mode)

    def load(self, order, model, data, clean_start=True, **kwargs):
        """
        Called by model new dataset is loaded.
        
        Here, Effects are configured from the data. self.parameters must be
        initialised after invoking this method.
        Parameters
        ----------
        order : int
            Identificator of effect in model. It is just an order of the effect
            among other effects as specified by user.
        model : ModelGeneralizedEffects
            Instance of ModelGeneralizedEffects that calls this method.
        data : pd.DataFrame
            Dataset that is being loaded. Should contain self.columns.
        clean_start : bool, optional
            If True, then parameters are (re)initialized. The model will use
            the ones already present in self.parameters vector otherwise. The
            default is True.

        Returns
        -------
        None.

        """
        super().load(order, model, data, clean_start, **kwargs)
        self.mx_identity = np.identity(data.shape[0])
        if clean_start:
            self.parameters = np.array([])

    def calc_k(self, model):
        return self.mx_identity

    def calc_k_grad(self, model):
        return []



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Regularization module."""
from functools import partial
import numpy as np


def create_regularization(model, regularization='l1-thresh', c=1.0, alpha=None,
                          param_names=None, mx_names=None):
    """
    Build additive regularization terms for loss funtions of SEM models.

    Supported regularizators:
        'l1-naive':  Straightforward L1 regularization (unrecommended).
        'l1-smooth': Smooth approximization of L1 regularization.
        'l1-thresh': L1 regularization with a soft thresholding operator.
        'l2-naive':  Straightforward L2 regularization
        'l2-square': L2 regularization squared.

    Parameters
    ----------
    model : Model
        Model reference.
    regularization : str, optional
        Type of of regularization to use. The default is 'l1-thresh'.
    c : TYPE, optional
        Regularization constant. The default is 1.0.
    alpha : float, optional
        If regularization is "l1-smooth", it's the alpha multiplier: the higher
        it's value is, the more accurate is the approximation. If
        regularization is "l1-thresh", then its a soft thresholding operator
        parameter: the closer to zero it is, the more accurate is the operator.
        The default is 1e-6.
    param_names : iterable, optional
        Parameter names to regularize. The default is None.
    mx_names : iterable, optional
        Parameters that are in mx_names will be regularized. The default is
        None.

    Returns
    -------
    Pair of objective and, if possible, a gradient function to be passed into
    Model.

    """
    inds = list()
    if param_names is None:
        param_names = set()
    if mx_names is None:
        mx_names = set()
    mx_names = [id(getattr(model, 'mx_{}'.format(mx.lower())))
                for mx in mx_names]
    i = 0
    for name, p in model.parameters.items():
        if not p.active:
            continue
        if name in param_names:
            inds.append(i)
        else:
            t = False
            for loc in p.locations:
                try:
                    if id(loc.matrix) in mx_names:
                        t = True
                        break
                except ValueError:
                    pass
            if t:
                inds.append(i)
        i += 1
    inds = np.array(inds, dtype=np.int)
    if regularization == 'l1-naive':
        obj = partial(l1_naive, c=c, inds=inds)
        grad = partial(l1_naive_grad, c=c, inds=inds)
    elif regularization == 'l1-smooth':
        if alpha is None:
            alpha = 1e6
        else:
            alpha = 1 / alpha
        obj = partial(l1_smooth, c=c, alpha=alpha, inds=inds)
        grad = partial(l1_smooth_grad, c=c, alpha=alpha, inds=inds)
    elif regularization == 'l1-thresh':
        if alpha is None:
            alpha = 1e-6
        obj = partial(l1_thresh, c=c, alpha=alpha, inds=inds)
        grad = partial(l1_thresh_grad, c=c, alpha=alpha, inds=inds)
    elif regularization == 'l2-naive':
        obj = partial(l2_naive, c=c, inds=inds)
        grad = partial(l2_naive_grad, c=c, inds=inds)
    elif regularization == 'l2-square':
        obj = partial(l2_square, c=c, inds=inds)
        grad = partial(l2_square_grad, c=c, inds=inds)
    else:
        raise NotImplementedError(f"Unknown regularization {regularization}.")
    return obj, grad


def l1_naive(x: np.ndarray, c: float, inds: np.ndarray):
    return c * np.abs(x[inds]).sum()


def l1_naive_grad(x: np.ndarray, c: float, inds: np.ndarray):
    g = np.zeros(len(x))
    for i in inds:
        g[i] = c * np.sign(x[i])
    return g


def l1_smooth(x: np.ndarray, c: float, alpha: float, inds: np.ndarray):
    t = alpha * x[inds]
    e_m = np.exp(-t)
    e_p = np.exp(t)
    r= (np.log(e_p + 1) + np.log(e_m + 1)).sum() / alpha
    return r


def l1_smooth_grad(x: np.ndarray, c: float, alpha: float, inds: np.ndarray):
    g = np.zeros(len(x))
    g[inds] = x[inds]
    g = np.tanh(alpha / 2 * g) 
    return g


def l1_thresh(x: np.ndarray, c: float, alpha: float, inds: np.ndarray):
    return c * np.abs(x[inds]).sum()


def l1_thresh_grad(x: np.ndarray, c: float, alpha: float, inds: np.ndarray):
    g = np.zeros(len(x))
    g[inds] = x[inds]
    t = g > alpha
    g[t] -= alpha
    t = g < -alpha
    g[t] += alpha
    t = np.abs(g) < alpha
    g[t] = 0.0
    return g


def l2_naive(x: np.ndarray, c: float, inds: np.ndarray):
    return c * np.sqrt(np.square(x[inds]).sum())


def l2_naive_grad(x: np.ndarray, c: float, inds: np.ndarray):
    g = np.zeros(len(x))
    t = x[inds]
    sqr = np.sqrt(np.square(t).sum())
    for i in inds:
        g[i] = x[i] / sqr
    return c * g


def l2_square(x: np.ndarray, c: float, inds: np.ndarray):
    return c * np.square(x[inds]).sum()


def l2_square_grad(x: np.ndarray, c: float, inds: np.ndarray):
    g = np.zeros(len(x))
    for i in inds:
        g[i] = 2 * c * x[i]
    return g



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import unittest
import numpy as np
import pandas as pd
from itertools import combinations
from ..model import Model
from ..model_means import ModelMeans
from ..model_effects import ModelEffects
from ..examples import multivariate_regression


class TestStandardErrors(unittest.TestCase):
    def evaluate(self, desc: str, data: pd.DataFrame):
        g = data['group'].copy()
        data = data - data.mean()
        data['group'] = g
        model = Model(desc)
        model_means = ModelMeans(desc, intercepts=False)
        model_effects = ModelEffects(desc, intercepts=False)
        assert model.fit(data).success, "Optimization failure."
        r = model_means.fit(data)
        if type(r) is tuple:
            assert r[0].success and r[1].success, "Optimization failure."
        else:
            assert r.success, "Optimization failure."
        r = model_effects.fit(data, group='group')
        if type(r) is tuple:
            assert r[0].success and r[1].success, "Optimization failure."
        else:
            assert r.success, "Optimization failure."
        se = list()
        ins = model.inspect(information='expected')
        ins = ins[ins['op'] == '~'].sort_values(['lval', 'rval'], 0)
        se.append(ins['Std. Err'].values)
        ins = model.inspect(information='observed')
        ins = ins[ins['op'] == '~'].sort_values(['lval', 'rval'], 0)
        se.append(ins['Std. Err'].values)
        ins = model_means.inspect(information='expected')
        ins = ins[ins['op'] == '~'].sort_values(['lval', 'rval'], 0)
        se.append(ins['Std. Err'].values)
        ins = model_means.inspect(information='observed')
        ins = ins[ins['op'] == '~'].sort_values(['lval', 'rval'], 0)
        se.append(ins['Std. Err'].values)
        ins = model_effects.inspect(information='expected')
        ins = ins[ins['op'] == '~'].sort_values(['lval', 'rval'], 0)
        se.append(ins['Std. Err'].values)
        ins = model_effects.inspect(information='observed')
        ins = ins[ins['op'] == '~'].sort_values(['lval', 'rval'], 0)
        se.append(ins['Std. Err'].values)
        for a, b in combinations(se, 2):
            t = (a < 1e-4) & (b < 1e-4)
            m = np.max(np.abs(a - b) / b * t)
            assert m < 1e-2, "Standard errors diverge {:.3f}.".format(m)

    def test_multivariate_regression(self):
        desc = multivariate_regression.get_model()
        data = multivariate_regression.get_data()
        data['group'] = 1
        self.evaluate(desc, data)



# -*- coding: utf-8 -*-
"""Random Effects models."""
from .effect_base import EffectBase
from .effect_blank import EffectBlank
from .effect_static import EffectStatic
from .effect_free import EffectFree
from .effect_ma import EffectMA
from .effect_ar import EffectAR
from .effect_matern import EffectMatern
from .effect_kernel import EffectKernel